program DP_test
 !!*** Perform dot-product tests
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
  use rtm_mod, only : rtm_image
  use born_mod, only : born_foo
  use wemva_mod, only : wemva_foo
  use taper_mod
  use skull_mod
  use assertions
  use born_op_mod
  use wemva_op_mod
!  use H_mod

  implicit none
  integer                 :: time(8), seed(2)
  integer :: jt, stat
  real*8 :: dpData, dpRefl, err_abs, err_rel
  real, allocatable :: rand_var(:,:)
  real, dimension(:), allocatable :: d1, d2, r1, r2, rv, dv, rmv, dmv 
  real, dimension(:,:), allocatable :: r, rm, d, dm
  real, dimension(:,:,:), allocatable :: dpxz
  logical, parameter :: born_write=F, test=F
  character :: op
  type(src_prop_fwd)     :: fwd   ! Arrays and procedures for source forward propagation
  type(src_rcv_prop_bwd) :: bwd   ! Arrays and procedures for source/receiver backward propagation
  type(born_foo)         :: born  ! Arrays and procedures for imaging conditions
  type(rtm_image)        :: image ! Arrays and procedures for imaging conditions
  type(wemva_foo)        :: wemva ! Arrays and procedures for imaging conditions
  type(read_)            :: par   ! Contains the input parameters and reading procedures
  type(write_)           :: outp  ! Contains the ouput parameters and reading procedures
  type(vshot_foo)        :: vel
  type(rshot_foo)        :: ref

  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()

  !! Choose dptest for Born modeling or WEMVA
  call from_param('op',op,'B')

  !! De-activate energy imaging conditions for dptest
  par%eic=T
  
  !! Test only in one shot
  par%one_src='y'
  
  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()

  !! Read reflectivity
  call par%read_reflectivity()

  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  
  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()
  
  !! Read and verbose source
  call par%read_src()
  
  !! Verbose some useful parameters
  call par%verbose_par()
  
  !! Initialize parameters for forward and backward propagations
  call fwd%fwd_prop_init(par)
  call bwd%bwd_prop_init(par)
  
  !! Initialize first derivatives parameters
!  call fwd%grad_init()
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize tapering for receiver propagation
!  call taper_init(par)
  
  !! Initialize derivatives coeffiecients
  call fwd%lap_init()
 
  !! Allocate arrays for the dot-product test 
  allocate(r(par%nkz,par%nkgx))
  allocate(rv(par%nkz*par%nkgx))
  allocate(rm(par%nkz,par%nkgx))
  allocate(rmv(par%nkz*par%nkgx))
  select case(op)
    case('W')
      allocate(d(par%nkz,par%nkgx))
      allocate(dv(par%nkz*par%nkgx))
      allocate(dm(par%nkz,par%nkgx))
      allocate(dmv(par%nkz*par%nkgx))
    case('B')
      allocate(d(par%nt,par%nkgx))
      allocate(dv(par%nt*par%nkgx))
      allocate(dm(par%nt,par%nkgx))
      allocate(dmv(par%nt*par%nkgx))
!    case('H')
!      allocate(d(par%nz,par%nx))
!      allocate(dv(par%nz*par%nx))
!      allocate(dm(par%nz,par%nx))
!      allocate(dmv(par%nz*par%nx))
    case default
      call skull()
      error stop "DP_test: Unknown operator"
  end select

  !! Produce random data for the dot-product test
  call date_and_time(values=time)     ! Get the current time 
  seed(1) = time(4) * (360000*time(5) + 6000*time(6) + 100*time(7) + time(8))
  call random_seed(put=seed)

  call random_number(rm)
  rm=rm-0.5

  call random_number(dm)
  dm=dm-0.5

  select case(op)
    case('W')
      !! Initialize parameters for forward operator
      call wemva%wemva_init(par,fwd,sx,gx)

      !! Initialize parameters for born inversion
      call wemva_op_init1(par,wemva)
    case('B')
      !! Initialize parameters for forward operator
      call born%born_init(par,fwd,sx,gx)
  
      !! Initialize parameters for adjoint operator
      call image%rtm_init(par,fwd,bwd,sx,gx)
  
      !! Initialize parameters for born inversion
      call born_op_init1(par,image,born)
!    case('H')
!      !! Initialize parameters for GN Hessian inversion
!      call H_int(par)
  end select
 
  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
  
  !! Initialize writing module array
  call outp%write_init(par) 

  if(test) goto 100

  !! This is the main loop to migrate the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshot,par%sgx(:,jsources), .false.)
    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),.false.)
    call sep_write(fwd%Vshot,'vrandom')
    if(.not. allocated(bwd%Vshot))  allocate(bwd%Vshot, mold=fwd%Vshot)
    bwd%Vshot=fwd%Vshot
    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,mold=fwd%Vshotr)
    bwd%Vshotr=fwd%Vshotr

    !! Compute shot position and report progress
    fwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    bwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof',  &
    par%nsources,jsources,fwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof
  
    !! Initialize updated velocity and wavefield classes
    select case(op)
      case('W')
        !! Reading data (traces from one shot gather)
        call par%read_shot()
        call wemva_op_init2(par%rcv,vel,ref,fwd,bwd,jsources)
      case('B')
        call born_op_init2(vel,ref,fwd,bwd,jsources)
!      case('H')
!        call H_init(par,T)       
      case default
        call skull()
        error stop "DP_test(1): Unknown operator"  
    end select

    !! Perform forward operator for each shot gather
    rmv=reshape(rm,(/size(rm,1)*size(rm,2)/))
    dv=reshape(d,(/size(d,1)*size(d,2)/))
    select case(op)
      case('W')
        stat=wemva_op(F,F,rmv,dv)
      case('B')
        stat=born_op(F,F,rmv,dv)
!      case('H')
!        stat=H_op(F,F,rmv,dv)
      case default 
        call skull()
        error stop "DP_test(2): Unknown operator"  
    end select
    d=reshape(dv,(/size(d,1),size(d,2)/))

    !! Perform adjoint operator in each shot gather
    dmv=reshape(dm,(/size(dm,1)*size(dm,2)/))
    rv=reshape(r,(/size(r,1)*size(r,2)/))
    select case(op)
      case('W')
        stat=wemva_op(T,F,rv,dmv)
      case('B')
        stat=born_op(T,F,rv,dmv)
!      case('H')
!        stat=H_op(T,F,rv,dmv)
      case default 
        call skull()
        error stop "DP_test(3): Unknown operator"  
    end select
    r=reshape(rv,(/size(r,1),size(r,2)/))

    !! Write the images history files and binaries
!    if(born_write) call outp%write_history(par,born,jsources)
  
    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  end do loop_over_sources
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''

  call assert(size(r)==size(rm),"DP_test: r and rm have different sizes!")
  call assert(size(d)==size(dm),"DP_test: d and dm have different sizes!")

  !! Perform the dot-product test
  write(0,*) 'Perform the dot-product test:'
  allocate(d1(size(d,1)*size(d,2)),&
           d2(size(d,1)*size(d,2)))
  allocate(r1(size(r,1)*size(r,2)),&
           r2(size(r,1)*size(r,2)))
  d1=reshape(d, (/size(d, 1)*size(d, 2)/))
  r1=reshape(r, (/size(r, 1)*size(r, 2)/))
  d2=reshape(dm,(/size(dm,1)*size(dm,2)/))
  r2=reshape(rm,(/size(rm,1)*size(rm,2)/))
  dpData=dot_product(d1,d2)
  dpRefl=dot_product(r1,r2)
  err_abs=dpData-dpRefl
  err_rel=err_abs/dpData
  write(0,*) 'dpData: ',dpData
  write(0,*) 'dpRefl: ',dpRefl
  write(0,*) 'err_abs: ',err_abs
  write(0,*) 'err_rel: ',err_rel
  deallocate(d1,d2,r1,r2)
  deallocate(d,r)

100 continue
  if(.not. test) goto 200

  !! This is a small local test
  !! Create velocity frames for source wavefield propagation
  call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),.false.)
  fwd%Vshot=fwd%Vshotr
  call sep_write(fwd%Vshot,'vrandom')
  if(.not. allocated(bwd%Vshot)) allocate(bwd%Vshot,mold=fwd%Vshot)
  if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,mold=fwd%Vshotr)
  bwd%Vshot=fwd%Vshot
  bwd%Vshotr=fwd%Vshotr
  fwd%pftm2=0
  fwd%pftm1=0
  fwd%pft=rm
  do jt=2,par%nt
    call fwd%fwd_prop(F)
  end do
  allocate(d,source=fwd%pft)
  bwd%prtp2=0
  bwd%prtp1=0
  bwd%prt=dm
  do jt=par%nt-1,1,-1
    call bwd%bwd_prop(fwd)
  end do
  allocate(r,source=bwd%prt)

!  allocate(dpxz(par%nkz,par%nkgx,2))
!  dpxz=0.
!  dpxz(:,:,1)=rm
!  call fwd%lap(dpxz,.true.,.false.)
!  allocate(d,source=dpxz(:,:,2))
!
!  dpxz=0.
!  dpxz(:,:,1)=dm
!  call fwd%lap(dpxz,.true.,.true.)
!  allocate(r,source=dpxz(:,:,2))

  write(0,*) 'Perform the local dot-product test:'
  !! Perform the dot-product test
  allocate(d1(size(d,1)*size(d,2)),&
           d2(size(d,1)*size(d,2)))
  allocate(r1(size(r,1)*size(r,2)),&
           r2(size(r,1)*size(r,2)))
  d1=reshape(d, (/size(d, 1)*size(d, 2)/))
  r1=reshape(r, (/size(r, 1)*size(r, 2)/))
  d2=reshape(dm,(/size(dm,1)*size(dm,2)/))
  r2=reshape(rm,(/size(rm,1)*size(rm,2)/))
  dpData=dot_product(d1,d2)
  dpRefl=dot_product(r1,r2)
  err_abs=dpData-dpRefl
  err_rel=err_abs/dpData
  write(0,*) 'dpData: ',dpData
  write(0,*) 'dpRefl: ',dpRefl
  write(0,*) 'err_abs: ',err_abs
  write(0,*) 'err_rel: ',err_rel

200 continue

  !! Write the images history files and binaries
!  if(.not. born_write) call outp%write_history(par,image,.false.)
!  call outp%write_history(par,image,wemva)

  
  !! Get the timing
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the DPtest program normally!'

end program
