submodule(H_mod) H_mod_implementation
  implicit none
contains

  module procedure H_init
    call sep_init() 
    call image%rtm_init(par,fwd,bwd,sx,gx)
    nz = par%nz
    nx = par%nx
    first_cdp=par%first_cdp
    last_cdp=par%last_cdp
    nlagz=par%nlagz
    nlagx=par%nlagx
    incPSFz=par%incPSFz
    incPSFx=par%incPSFx

    nzc=ceiling(1.*nz/incPSFz)
    nxc=ceiling(1.*nx/incPSFx)
  
    allocate(Hgn(nz,par%first_cdp:par%last_cdp,-nlagz:nlagz,-nlagx:nlagx))

    associate(x1=>par%first_cdp,x2=>par%last_cdp)
      allocate(m(nzc,nxc),             &
               d(par%nz*par%nx), &
               temp(par%nz,x1:x2),  &
               diff(par%nz,x1:x2),  &
               xz(par%nz*par%nx,2),  &
               stat=allocation_status)
      if(allocation_status /= 0) then
        write(0,*) 'Hess2d: Cannot allocate arrays for Hessian!'
        call skull()
      end if
        call sep_read(temp,"psf")
      do jx=x1,x2
        x=(jx-1)*par%dx
        do jz=1,par%nz
          z=(jz-1)*par%dz
          j=(jx-x1)*par%nz+jz
          xz(j,1)=z
          xz(j,2)=x
        end do
      end do
      d1=incPSFz*par%dz
      d2=incPSFx*par%dx
      o1=incPSFz*par%dz
      o2=(x1+incPSFx-1)*par%dx
    end associate

    do jlagx = -nlagx,nlagx
      do jlagz = -nlagz,nlagz
  
        !! Extrac elements of the GN Hessian
        associate(x1=>par%first_cdp,x2=>par%last_cdp)
          jxc=0
          m=0.
          do jx=x1+incPSFx,x2,incPSFx
            jxc=jxc+1
            jzc=0
            do jz=1+incPSFz,par%nz,incPSFz
              jzc=jzc+1
              m(jzc,jxc)=temp(jz+jlagz,jx+jlagx)
            end do
          end do
          m(nzc,:)=m(nzc-1,:)
          m(:,nxc)=m(:,nxc-1)
        end associate

        !! Interpolate the diagonal of the Hessian
        d=0.
        call lint2_init(nzc,nxc,o1,d1,o2,d2,xz)
        call lint2_lop2(F,T,m,d)

        !! Smooth the diagonal of the Hessian
        temp=reshape(d,(/par%nz,par%nx/))
        dummy=base_tris2d_oper(temp,20,20,1) 
        Hgn(:,:,jlagz,jlagx)=temp
      end do
    end do

    !! Extract the diagonal of the Hessian
    if(present(diagH)) diagH=Hgn(:,:,0,0)
  
    !! Write the interpolated Hessian
    associate(nz=>par%nz,x1=>par%first_cdp,x2=>par%last_cdp)
      write(0,*) 'Write the interpolated Hessian'
      call outp%write_history(par,image,Hgn,nz,x1,x2,nlagz,nlagx)
    end associate
  end procedure

  module procedure H_op
    call H_op2(adj, add, model, data_)
    stat=0
  end procedure

  module procedure H_op2
    if(adj) then
      if(.not. add) model = 0.
      do jlagx = -nlagx,nlagx
        do jlagz = -nlagz,nlagz
          do jx = 1,nx
            do jz = 1,nz
              model(jz,jx)=model(jz,jx)&
                +Hgn(jz,jx+first_cdp-1,jlagz,jlagx)*data_(jz+jlagz,jx+jlagx)
            end do
          end do
        end do
      end do
    else
      if(.not. add) data_ = 0.
      do jlagx = -nlagx,nlagx
        do jlagz = -nlagz,nlagz
          do jx = 1,nx
            do jz = 1,nz
              data_(jz+jlagz,jx+jlagx)=data_(jz+jlagz,jx+jlagx)&
                +Hgn(jz,jx+first_cdp-1,jlagz,jlagx)*model(jz,jx)
            end do
          end do
        end do
      end do
    end if

  end subroutine

end submodule
