submodule(born_mod) born_mod_implementation
  implicit none
contains
  module procedure born
    !! Create velocity and reflect. frames for source scatt. wavefield propagation
    if(present(temp)) then
      call vel%vshot(par, scat%Vshot,   par%sgx(:,jsources),.false.)
    else
      call vel%vshot(par, scat%Vshot,   par%sgx(:,jsources),.true.)
    end if

    if(.not. present(temp)) call ref%rshot(par,this%reflshot,par%sgx(:,jsources))
    !! Initialize/update wavefield arrays
    call  fwd%init_src_wavefields()
    call scat%init_src_wavefields()
  
    !! Source scaling factors (using dipole sources)
    associate(z1=>par%zpad+par%pad_,x1=>par%pad_+par%napL+fwd%shot_pos)
      src_scl1=-(fwd%Vshot(1+z1,x1)*par%dt)**2
      src_scl2=-(fwd%Vshot(2+z1,x1)*par%dt)**2
    end associate
    if(.not. allocated(scat_scl)) allocate(scat_scl(par%nkz,par%nkgx))
    scat_scl=-(fwd%Vshot(:,:)*par%dt)**2

    !! Initialize synthetic data and scattered wavefield arrays
    this%dsynt=0

!DELETE AFTER TEST
    if(present(temp)) this%reflshot=temp
    call sep_write(this%reflshot,'rrandom')

    write(0,*) 'Perform born modelling'

    do jt=1,par%nt
      !! Forward propagation of source wavefield
      if(jt>1) call fwd%fwd_prop(F)   

      !! Inject the source
      fwd%pft(1,fwd%shot_pos)=fwd%pft(1,fwd%shot_pos)+par%src(jt)*src_scl1
      fwd%pft(2,fwd%shot_pos)=fwd%pft(2,fwd%shot_pos)-par%src(jt)*src_scl2

      !! Damping function applied to the source wavefield
      if(use_taper) call taperings(fwd%pftm2,fwd%pftm1,fwd%pft)
!      if(use_taper) call taperings(fwd%d2pftm2,fwd%d2pftm1,fwd%d2pft)

      !! Obtain snapshots of the source wavefield for a single source
!      if(par%snapsht .and. jsources == par%src_snap&
!                     .and. mod(jt,jtsnap) == 0) then
!        if(.not. present(temp)) par%ntSnap=par%ntSnap+1
!        call fwd%get_snaps_src(par%sgx(:,jsources))
!      end if

      !! Scatter the source wavefield upon the reflectivity
      scat%pft=scat%pft+fwd%pft*this%reflshot*scat_scl

      !! Propagate the scattered  source wavefield
      call scat%fwd_prop(F)   

      !! Damping function applied to the scattered wavefield
      if(use_taper) call taperings(scat%pftm2,scat%pftm1,scat%pft)
!      if(use_taper) call taperings(scat%d2pftm2,scat%d2pftm1,scat%d2pft)

      !! Obtain snapshots of the source wavefield for a single source
      if(par%snapsht .and. jsources == par%src_snap .and. mod(jt,jtsnap) == 0) &
        call scat%get_snaps_scat(par%sgx(:,jsources))

      !! Extract the wavefield on the surface
      this%dsynt(jt,:)=this%dsynt(jt,:)+scat%pft(1,:)
      this%dsynt(jt,:)=this%dsynt(jt,:)-scat%pft(2,:)
    end do
  end procedure

  module procedure born_init
    !! Group some useful image limits
    x1 = 1-par%napL
    x2 = par%ngx+par%napR
    z1 = 1-par%zpad-par%pad_
    z2 = par%nz+par%pad_

    !! Allocate arrays for scattered wavefield computations 
    if(.not. allocated(this%reflshot)) then
      allocate(this%reflshot,mold=fwd%Vshot, stat=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'born_init: could not perform mold allocation of reflshot!'
      end if
    end if

    allocate(this%dsynt(par%nt,x1-par%pad_:x2+par%pad_),stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'born_init: could not allocate dsynt array!'
    end if
 
    allocate(scat,source=fwd,stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'born_init: could not assign parameter object "scat"'
    end if

!    allocate(owt(par%ngx),stat=allocation_status)
!    if(allocation_status /= 0) then
!      call skull()
!      error stop 'born_init: could not allocate owt array!'
!    end if
!
!    ! This is a tapering function with offset, employed by Ali (disabled now)
!    do jgx=1,par%ngx
!      owt(jgx) = -1.*abs(sx(par%trace_shot)-gx(jgx))
!    end do
!    owt = owt - minval(owt) + 1.
!    owt = sin(pi/2.*owt/maxval(owt))*sin(pi/2.*owt/maxval(owt)) 
  end procedure

end submodule
