module bwd_propagation_mod
  !! This module contains all procedures for computing 
  !! backward propagation of the source wavefield and the receiver wavefield
  use sep
  use fwd_propagation_mod, only : src_prop_fwd
!  use lap_c_mod
  use interp1d_mod
  use skull_mod
  use read_mod, only : read_
!  use Laplacian
  implicit none
  private
  public :: src_rcv_prop_bwd

  integer :: allocation_status
  integer :: jz, jgx, jx, jpol
  integer, dimension(:), allocatable :: encod_vector
  
  real, parameter :: pi = 4.*atan(1.) ! The Order of Chaos (Darren Aronofsky, 1998)
  real :: fx, gx, dt1, dt2, dt3, dt4
  real, dimension(:,:),   allocatable :: psnap
  real, dimension(:,:,:), allocatable :: dpxz, drxz

  type(read_), allocatable :: par

  type, extends(src_prop_fwd) :: src_rcv_prop_bwd
    !! This class extends object src_prop_fwd and add receiver wavefields
    real, dimension(:,:), allocatable :: prtp1, prtp2, prt, prtX, prtZ 
    real, dimension(:,:), allocatable :: dprtp1, dprtp2, dprt
    real, dimension(:,:), allocatable :: d2prtp1, d2prtp2, d2prt
  contains
    procedure init_rcv_wavefields
    procedure bwd_prop_init
    procedure grad_rcv
    procedure update_wavefields
    procedure get_snaps_rcv
    procedure bwd_prop
  end type

  interface
    pure module subroutine init_rcv_wavefields(this,fwd)
      !! This procedure passes the values of the source-wavefield-related arrays
      !! from class src_prop_fwd, and sets to zero receiver-wavefield related arrays
      implicit none
      class(src_rcv_prop_bwd), intent(inout) :: this
      type(src_prop_fwd), intent(in)      :: fwd
    end subroutine

    pure module subroutine bwd_prop_init(this,par_in)
      !! This procedure initializes all the scalar variables
      implicit none
      class(src_rcv_prop_bwd), intent(inout) :: this
      class(read_), intent(in) :: par_in
    end subroutine

    pure module subroutine grad_rcv(this,fwd)
      !! This procedure applies the gradient to the receiver wavefield
      implicit none
      class(src_rcv_prop_bwd), intent(inout) :: this
      type(src_prop_fwd), intent(inout)      :: fwd
    end subroutine

    pure module subroutine update_wavefields(this,fwd)
      !! This procedure passes the source wavefield to the reversible version 
      implicit none
      class(src_rcv_prop_bwd), intent(inout) :: this
      type(src_prop_fwd), intent(in)      :: fwd
    end subroutine
  
    pure module subroutine get_snaps_rcv(this,sgx)
      !! This procedure obtains snapshots of the reciver wavefield from the chosen shot gather
      implicit none
      class(src_rcv_prop_bwd), intent(in) :: this
!      integer, intent(in) :: src_snap
      real, intent(in) :: sgx(:)
    end subroutine

    pure module subroutine bwd_prop(this,fwd)
      !! This procedure performs the backward propagation using the Lax-Wendroff method
      implicit none
      class(src_rcv_prop_bwd), intent(inout) :: this
      type(src_prop_fwd),intent(in)      :: fwd
    end subroutine
  end interface
end module
