submodule(bwd_propagation_mod) bwd_propagation_mod_implementation
  implicit none
contains

  module procedure init_rcv_wavefields
    !! Initialize source wavefield variables
    this%prt    =0.
    this%prtp1  =0.
    this%dprt   =0.
    this%dprtp1 =0.
    this%d2prt  =0.
    this%d2prtp1=0.
    this%pft    =fwd%pft 
    this%pftm1  =fwd%pftm1
    this%dpft   =fwd%dpft 
    this%dpftm1 =fwd%dpftm1
    this%d2pft  =fwd%d2pft 
    this%d2pftm1=fwd%d2pftm1
    this%Vshot  =fwd%Vshot
    this%Vshotr =fwd%Vshotr
  end procedure

  module procedure bwd_prop_init
    allocate(par,source=par_in)
    associate(nkz=>par%nkz,nkgx=>par%nkgx,ngx=>par%ngx,nt=>par%nt,nz=>par%nz,nx=>par%nx)
    associate(pad_=>par%pad_,zpad=>par%zpad,src_pad=>par%src_pad,dt=>par%dt)
    associate(z1=>1-pad_-zpad,z2=>nz+pad_,x1=>1-pad_-par%napL,x2=>ngx+par%napR+pad_)
    associate(x3=>par%first_cdp-par%pad_,x4=>par%last_cdp+par%pad_)
      allocate(dpxz(nkz,nkgx,2),&
               psnap(z1:z2,x3:x4),      &
               stat=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'fwd_prop_init: could not allocate arrays!'
      end if
      !! Allocate arrays for the main calculations
      allocate(this%prt    (z1:z2,x1:x2+src_pad),&
               this%prtp1  (z1:z2,x1:x2+src_pad),&
               this%prtp2  (z1:z2,x1:x2+src_pad),&
               this%dprt   (z1:z2,x1:x2+src_pad),&
               this%dprtp1 (z1:z2,x1:x2+src_pad),&
               this%dprtp2 (z1:z2,x1:x2+src_pad),&
               this%d2prt  (z1:z2,x1:x2+src_pad),&
               this%d2prtp1(z1:z2,x1:x2+src_pad),&
               this%d2prtp2(z1:z2,x1:x2+src_pad),&
               this%prtX   (z1:z2,x1:x2+src_pad),&
               this%prtZ   (z1:z2,x1:x2+src_pad),&
               this%pft    (z1:z2,x1:x2+src_pad),&
               this%pftm1  (z1:z2,x1:x2+src_pad),&
               this%pftm2  (z1:z2,x1:x2+src_pad),&
               this%dpft   (z1:z2,x1:x2+src_pad),&
               this%dpftm1 (z1:z2,x1:x2+src_pad),&
               this%dpftm2 (z1:z2,x1:x2+src_pad),&
               this%d2pft  (z1:z2,x1:x2+src_pad),&
               this%d2pftm1(z1:z2,x1:x2+src_pad),&
               this%d2pftm2(z1:z2,x1:x2+src_pad),&
               this%pftX   (z1:z2,x1:x2+src_pad),&
               this%pftZ   (z1:z2,x1:x2+src_pad),&
               this%src_scl(nt),                 &
               this%rcv_scl(nt,ngx),             &
               STAT=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'bwd_prop_init: could not allocate RTM arrays!'
      end if
      dt1=par%dt
      dt2=dt1*dt1
      dt3=dt2*dt1
      dt4=dt2*dt2
    end associate
    end associate
    end associate
    end associate
  end procedure

  module procedure grad_rcv
    fwd%pft=this%prt
    call fwd%grad()
    this%prtX=fwd%pftX
    this%prtZ=fwd%pftZ
    call this%grad()
  end procedure

  module procedure update_wavefields
    this%prtp1=0.
    this%prt=0.
    this%dprtp1=0.
    this%dprt=0.
    this%d2prtp1=0.
    this%d2prt=0.
    this%pftm1  =fwd%pft
    this%pft    =fwd%pftm1
    this%dpftm1 =fwd%dpft
    this%dpft   =fwd%dpftm1
    this%d2pftm1=fwd%d2pft
    this%d2pft  =fwd%d2pftm1
  end procedure

  module procedure get_snaps_rcv
    associate(x1=>par%first_cdp-par%pad_,x2=>par%last_cdp+par%pad_)
      do jgx = 1-par%napL,par%ngx+par%napR
!        jx = nint(1.5001+sgx(jgx+par%napL,par%src_snap)/par%dx)
        jx = nint(1.5001+sgx(jgx+par%napL)/par%dx)
        if(jx >= x1 .and. jx <= x2) then
          psnap(:,jx) = this%pft(:,jgx)
        end if
      end do
      do jx = x1,x2
        call sep_write(psnap(:,jx),"snapRcv")
      end do
    end associate
  end procedure

  module procedure bwd_prop
    !! Perform receiver propagation backwards
    this%prtp2   = this%prtp1
    this%prtp1   = this%prt
    this%dprtp2  = this%dprtp1
    this%dprtp1  = this%dprt
    this%d2prtp2 = this%d2prtp1
    this%d2prtp1 = this%d2prt
    dpxz = 0
    dpxz(:,:,1) = this%prtp1
    call fwd%lap(dpxz,.false.,.true.)
    this%prt = this%prtp1*2.-this%prtp2+dpxz(:,:,1)*dt2+dpxz(:,:,2)*dt4/12.
    this%dprt = this%dprtp2-2.*dt1*dpxz(:,:,1)-dt3*dpxz(:,:,2)/3.
    this%d2prt = -this%d2prtp2+2.*dpxz(:,:,1)+dt2*dpxz(:,:,2)!+dpxz(:,:,3)*dt4/12.

    !! Perform source propagation backwards
    this%pftm2   = this%pftm1
    this%pftm1   = this%pft
    this%dpftm2  = this%dpftm1
    this%dpftm1  = this%dpft
    this%d2pftm2 = this%d2pftm1
    this%d2pftm1 = this%d2pft
    dpxz = 0
    dpxz(:,:,1) = this%pftm1
    call fwd%lap(dpxz,.true.,.false.)
    this%pft = this%pftm1*2.-this%pftm2+dpxz(:,:,1)*dt2+dpxz(:,:,2)*dt4/12.
    this%dpft = this%dpftm2-2.*dt1*dpxz(:,:,1)-dt3*dpxz(:,:,2)/3.
    this%d2pft = -this%d2pftm2+2.*dpxz(:,:,1)+dt2*dpxz(:,:,2)!+dpxz(:,:,3)*dt4/12.

  end procedure
end submodule
