submodule(fwd_propagation_mod) fwd_propagation_mod_implementation
  implicit none
contains
  module procedure fwd_prop_init
    if(.not. allocated(par)) allocate(par,source=par_in)
    associate(nkz=>par%nkz,nkgx=>par%nkgx,ngx=>par%ngx,nt=>par%nt,nz=>par%nz,nx=>par%nx)
    associate(pad_=>par%pad_,zpad=>par%zpad,src_pad=>par%src_pad)
    associate(z1=>1-pad_-zpad,z2=>nz+pad_,x1=>1-pad_-par%napL,x2=>ngx+par%napR+pad_)
    associate(x3=>par%first_cdp-par%pad_,x4=>par%last_cdp+par%pad_)
      if(.not. allocated(dpxz)) then
        allocate(dpxz(nkz,nkgx,2), &
                 psnap(z1:z2,x3:x4),      &
                 stat=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'fwd_prop_init: could not allocate arrays!'
        end if
      end if

      !! Allocate arrays for the main calculations
      if(.not. allocated(this%pft)) then
        allocate(this%pft    (z1:z2,x1:x2+src_pad),&
                 this%pftm1  (z1:z2,x1:x2+src_pad),&
                 this%pftm2  (z1:z2,x1:x2+src_pad),&
                 this%dpft   (z1:z2,x1:x2+src_pad),&
                 this%dpftm1 (z1:z2,x1:x2+src_pad),&
                 this%dpftm2 (z1:z2,x1:x2+src_pad),&
                 this%d2pft  (z1:z2,x1:x2+src_pad),&
                 this%d2pftm1(z1:z2,x1:x2+src_pad),&
                 this%d2pftm2(z1:z2,x1:x2+src_pad),&
                 this%pftX   (z1:z2,x1:x2+src_pad),&
                 this%pftZ   (z1:z2,x1:x2+src_pad),&
                 this%src_scl(nt),                 &
                 this%rcv_scl(nt,x1:x2+src_pad),   &
                 STAT=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'fwd_prop_init: could not allocate prop. arrays!'
        end if
      end if
    dt1=par%dt
    dt2=dt1*dt1
    dt3=dt2*dt1
    dt4=dt2*dt2
    end associate
    end associate
    end associate
    end associate
  end procedure

  module procedure grad_init
    this%use_grad=.true.
    associate(nkz=>par%nkz,nkgx=>par%nkgx,nz=>par%nz,nx=>par%nx)
    associate(dz=>par%dz,dx=>par%dx)
      idx = 1.0/dx
      idz = 1.0/dz
      if(.not. allocated(dpx_cur)) then  
        allocate(dpx_cur(1-buff:nkz+buff,1-buff:nkgx+buff),&
                 dpz_cur(1-buff:nkz+buff,1-buff:nkgx+buff),&
                 pauxgrad(max(nkz,nkgx)),&
                 pxcur(nkz*nkgx),&
                 rzcur(nkz*nkgx),&
                 stat=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'grad_init: could not allocate arrays!'
        end if
      end if
    end associate
    end associate

    !! 1st deriv, 12th order approx. table by Fornberg and Ghrist (1999)
    b1x = 160083.0/131072.0*idx
    b2x = -12705.0/131072.0*idx
    b3x = 22869.0/1310720.0*idx
    b4x = -5445.0/1835008.0*idx
    b5x =   847.0/2359296.0*idx
    b6x =   -63.0/2883584.0*idx

    b1z = 160083.0/131072.0*idz
    b2z = -12705.0/131072.0*idz
    b3z = 22869.0/1310720.0*idz
    b4z = -5445.0/1835008.0*idz
    b5z =   847.0/2359296.0*idz
    b6z =   -63.0/2883584.0*idz
  end procedure

  module procedure lap_init
    associate(nkz=>par%nkz,nkgx=>par%nkgx,nz=>par%nz,nx=>par%nx)
    associate(dz=>par%dz,dx=>par%dx)
      idx = 1.0/(dx*dx)
      idz = 1.0/(dz*dz)
 
      if(.not. allocated(this%Vshot)) then 
        allocate(this%Vshot(nkz,nkgx), &
                 this%Vshotr(nkz,nkgx),&
                 stat=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'lap_init: could not allocate Vshot arrays!'
        end if
      end if

      if(.not. allocated(dpxz_cur)) then 
        allocate(Vel2(nkz,nkgx),&
                 dpxz_cur(1-buff:nkz+buff,1-buff:nkgx+buff),&
                 stat=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'lap_init: could not allocate Vel2 and dpxz_cur arrays!'
        end if
      end if
    end associate
    end associate

   !! 2nd deriv., 12th order approx. Zhang and Yao (2013)
!    a0z = -3.12513824*idz
!    a1z =  1.84108651*idz
!    a2z = -0.35706478*idz
!    a3z =  0.10185626*idz
!    a4z = -0.02924772*idz
!    a5z =  0.00696837*idz
!    a6z = -0.00102952*idz

!    a0x = -3.12513824*idx 
!    a1x =  1.84108651*idx
!    a2x = -0.35706478*idx
!    a3x =  0.10185626*idx
!    a4x = -0.02924772*idx
!    a5x =  0.00696837*idx
!    a6x = -0.00102952*idx
!    a0 = a0x + a0z

    !! 2nd deriv., 10th order approx., courtesy of Ettore Biondi
    a0z = -2.927222222222182   *idz 
    a1z =  1.666666666666642   *idz
    a2z = -0.238095238095235   *idz
    a3z =  0.039682539682539   *idz
    a4z = -0.004960317460317   *idz
    a5z =  3.174603174603100e-4*idz
    
    a0x = -2.927222222222182   *idx  
    a1x =  1.666666666666642   *idx
    a2x = -0.238095238095235   *idx
    a3x =  0.039682539682539   *idx
    a4x = -0.004960317460317   *idx
    a5x =  3.174603174603100e-4*idx
    a0 = a0x + a0z
  end procedure

  module procedure grad
    !! I NEED TO TAKE CARE OF THIS PROCEDURE FOR ADJOINTNESS PURPOSES!
    dpx_cur=0.
    dpz_cur=0.
    dpx_cur(1:par%nkz,1:par%nkgx) = this%pft!X
    dpz_cur(1:par%nkz,1:par%nkgx) = this%pft!Z
    this%pftX = 0
    this%pftZ = 0
 !   Vel2=this%Vshotr**2
    !$omp parallel do default(shared) private(jgx,jz,jaux) 
    do jgx = 1,par%nkgx
      !$omp simd
      do jz = 1,par%nkz
        jaux=(jgx-1)*par%nkz+jz
        pxcur(jaux) = b1x*(dpx_cur(jz,jgx+1)-dpx_cur(jz,jgx  ))+&
                      b2x*(dpx_cur(jz,jgx+2)-dpx_cur(jz,jgx-1))+&
                      b3x*(dpx_cur(jz,jgx+3)-dpx_cur(jz,jgx-2))+&
                      b4x*(dpx_cur(jz,jgx+4)-dpx_cur(jz,jgx-3))+&
                      b5x*(dpx_cur(jz,jgx+5)-dpx_cur(jz,jgx-4))+&
                      b6x*(dpx_cur(jz,jgx+6)-dpx_cur(jz,jgx-5))
      end do
    end do
    !$omp end parallel do
    
    !$omp parallel do default(shared) private(jgx,jz,jaux) 
    do jgx = 1,par%nkgx
      !$omp simd
      do jz = 1,par%nkz
        jaux=(jgx-1)*par%nkz+jz
        rzcur(jaux) = b1z*(dpz_cur(jz+1,jgx)-dpz_cur(jz,jgx  ))+&
                      b2z*(dpz_cur(jz+2,jgx)-dpz_cur(jz-1,jgx))+&
                      b3z*(dpz_cur(jz+3,jgx)-dpz_cur(jz-2,jgx))+&
                      b4z*(dpz_cur(jz+4,jgx)-dpz_cur(jz-3,jgx))+&
                      b5z*(dpz_cur(jz+5,jgx)-dpz_cur(jz-4,jgx))+&
                      b6z*(dpz_cur(jz+6,jgx)-dpz_cur(jz-5,jgx))
      end do
    end do
    !$omp end parallel do
    this%pftX = reshape(pxcur,(/par%nkz,par%nkgx/)) 
    this%pftZ = reshape(rzcur,(/par%nkz,par%nkgx/)) 
    pauxgrad=0
    !$omp parallel do default(shared) private(pauxgrad,jgx) 
    do jgx = 1,par%nkgx
      call interp1d_lop2(.false.,.false.,this%pftZ(:,jgx),pauxgrad(1:par%nkz))
      this%pftZ(:,jgx)=pauxgrad(1:par%nkz)
    end do
    !$omp end parallel do

    pauxgrad=0
    !$omp parallel do default(shared) private(pauxgrad,jz) 
    do jz = 1,par%nkz
      call interp1d_lop2(.false.,.false.,this%pftX(jz,:),pauxgrad(1:par%nkgx))
      this%pftX(jz,:)=pauxgrad(1:par%nkgx)
    end do
    !$omp end parallel do
  end procedure

  module procedure lap
    if(isSrc) then
      Vel2=this%Vshot**2
    else
      Vel2=this%Vshotr**2
    end if
    do jpol = 1,2
      dpxz_cur=0.
      dpxz_cur(1:par%nkz,1:par%nkgx) = dpxz(:,:,jpol)
      dpxz(:,:,jpol)=0.

      !$omp parallel do default(shared) private(jgx,jz)
      if(adj) then
      dpxz_cur(1:par%nkz,1:par%nkgx) = dpxz_cur(1:par%nkz,1:par%nkgx)*Vel2
        do jgx = 1,par%nkgx
          !$omp simd
          do jz = 1,par%nkz
            dpxz(jz,jgx,jpol) = &
              a0 * dpxz_cur(jz,jgx)+  &
              a1x*(dpxz_cur(jz,jgx+1)+dpxz_cur(jz,jgx-1))+&
              a2x*(dpxz_cur(jz,jgx+2)+dpxz_cur(jz,jgx-2))+&
              a3x*(dpxz_cur(jz,jgx+3)+dpxz_cur(jz,jgx-3))+&
              a4x*(dpxz_cur(jz,jgx+4)+dpxz_cur(jz,jgx-4))+&
              a5x*(dpxz_cur(jz,jgx+5)+dpxz_cur(jz,jgx-5))+&
              a1z*(dpxz_cur(jz+1,jgx)+dpxz_cur(jz-1,jgx))+&
              a2z*(dpxz_cur(jz+2,jgx)+dpxz_cur(jz-2,jgx))+&
              a3z*(dpxz_cur(jz+3,jgx)+dpxz_cur(jz-3,jgx))+&
              a4z*(dpxz_cur(jz+4,jgx)+dpxz_cur(jz-4,jgx))+&
              a5z*(dpxz_cur(jz+5,jgx)+dpxz_cur(jz-5,jgx))
          end do
        end do
        !$omp end parallel do
      else
        do jgx = 1,par%nkgx
          !$omp simd
          do jz = 1,par%nkz
            dpxz(jz,jgx,jpol) = Vel2(jz,jgx)*(&
              a0 * dpxz_cur(jz,jgx)+  &
              a1x*(dpxz_cur(jz,jgx+1)+dpxz_cur(jz,jgx-1))+&
              a2x*(dpxz_cur(jz,jgx+2)+dpxz_cur(jz,jgx-2))+&
              a3x*(dpxz_cur(jz,jgx+3)+dpxz_cur(jz,jgx-3))+&
              a4x*(dpxz_cur(jz,jgx+4)+dpxz_cur(jz,jgx-4))+&
              a5x*(dpxz_cur(jz,jgx+5)+dpxz_cur(jz,jgx-5))+&
              a1z*(dpxz_cur(jz+1,jgx)+dpxz_cur(jz-1,jgx))+&
              a2z*(dpxz_cur(jz+2,jgx)+dpxz_cur(jz-2,jgx))+&
              a3z*(dpxz_cur(jz+3,jgx)+dpxz_cur(jz-3,jgx))+&
              a4z*(dpxz_cur(jz+4,jgx)+dpxz_cur(jz-4,jgx))+&
              a5z*(dpxz_cur(jz+5,jgx)+dpxz_cur(jz-5,jgx)))
          end do
        end do
        !$omp end parallel do
      end if
      if(jpol == 2) exit
      dpxz(:,:,jpol+1) = dpxz(:,:,jpol)
    end do
  end procedure

  module procedure init_src_wavefields
    !! Initialize source wavefield arrays
    this%pft    = 0
    this%pftm1  = 0
    this%dpftm1 = 0
    this%dpft   = 0
    this%d2pftm1 = 0
    this%d2pft = 0
  end procedure

  module procedure get_snaps_src
    associate(x1=>par%first_cdp-par%pad_,x2=>par%last_cdp+par%pad_)
      do jgx = 1-par%napL,par%ngx+par%napR
        jx = nint(1.5001+sgx(jgx+par%napL)/par%dx)
        if(jx >= x1 .and. jx <= x2) then
          psnap(:,jx) = this%pft(:,jgx)
        end if
      end do
      do jx = x1,x2
        call sep_write(psnap(:,jx),'snapSrc')
      end do
    end associate
  end procedure

  module procedure get_snaps_scat
    associate(x1=>par%first_cdp-par%pad_,x2=>par%last_cdp+par%pad_)
      do jgx = 1-par%napL,par%ngx+par%napR
        jx = nint(1.5001+sgx(jgx+par%napL)/par%dx)
        if(jx >= x1 .and. jx <= x2) then
          psnap(:,jx) = this%pft(:,jgx)
        end if
      end do
      do jx = x1,x2
        call sep_write(psnap(:,jx),'snapScat')
      end do
    end associate
  end procedure

  module procedure fwd_prop
    !! Perform source propagation forwards
    this%pftm2   = this%pftm1
    this%pftm1   = this%pft
    this%dpftm2  = this%dpftm1
    this%dpftm1  = this%dpft
    this%d2pftm2 = this%d2pftm1
    this%d2pftm1 = this%d2pft
    dpxz = 0
    dpxz(:,:,1) = this%pftm1
    call this%lap(dpxz,.true.,adj)
    this%pft = this%pftm1*2.-this%pftm2+dpxz(:,:,1)*dt2+dpxz(:,:,2)*dt4/12.
    this%dpft = this%dpftm2+2.*dt1*dpxz(:,:,1)+dt3*dpxz(:,:,2)/3.
    this%d2pft = -this%d2pftm2+2.*dpxz(:,:,1)+dt2*dpxz(:,:,2)!+dpxz(:,:,3)*dt4/12.
  end procedure
end submodule
