module nonlin_mod
  !! This module implements conventional correlation imaging conditions and
  !! energy imaging conditions, and build the migrated images
  use sep
  use declare_variables_mod, only : use_taper
  use read_mod, only : read_
  use fwd_propagation_mod, only : src_prop_fwd
!  use omp_lib
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  private
  public :: nonlin_foo

  integer :: dummy, allocation_status
  integer :: k, jt, jtCoarse, jx, jz, jgx, x1, x2
  integer, parameter :: jtsnap=10 
  real :: time0, time1
  real :: fx, jfx, src_scl1, src_scl2
  real, allocatable, dimension(:)   :: owt
  real, parameter :: pi = 4.*atan(1.) ! The Order of Chaos (Darren Aronofsky, 1998)
  logical, parameter :: T=.true., F=.false.

  type nonlin_foo
    !! This class contains the array for syntetic data 
    !! and  procedures for non-linear modeling
    real, dimension(:,:), allocatable :: dsynt
  contains
    procedure :: nonlin
    procedure :: nonlin_init
  end type

  interface
    pure module subroutine nonlin(this,par,fwd,vel,jsources)
      !! I apply non-linear modeling procedure here
      implicit none
      class(nonlin_foo), intent(inout) :: this
      type(src_prop_fwd),     intent(inout) :: fwd 
      type(read_), intent(inout) :: par
      type(vshot_foo), intent(inout) :: vel
      integer, intent(in) :: jsources
    end subroutine

    pure module subroutine nonlin_init(this,par,fwd,sx,gx)
      !! Here the scalar variables are initialized
      implicit none
      class(nonlin_foo), intent(inout) :: this
      class(src_prop_fwd), intent(inout) :: fwd
      type(read_), intent(in) :: par
      real, dimension(:), intent(in) :: sx, gx
    end subroutine
  end interface
end module
