submodule(nonlin_mod) nonlin_mod_implementation
  implicit none
contains
  module procedure nonlin
    !! Initialize/update wavefield arrays
    call  fwd%init_src_wavefields()
  
    !! Source scaling factors (using dipole sources)
    associate(z1=>par%zpad+par%pad_,x1=>par%pad_+par%napL+fwd%shot_pos)
      src_scl1=-(fwd%Vshot(1+z1,x1)*par%dt)**2
      src_scl2=-(fwd%Vshot(2+z1,x1)*par%dt)**2
    end associate

    !! Initialize synthetic data
    this%dsynt=0

    do jt=1,par%nt
      !! Forward propagation of source wavefield
      call fwd%fwd_prop(F)   

      !! Inject the source
      fwd%pft(1,fwd%shot_pos)=fwd%pft(1,fwd%shot_pos)+par%src(jt)*src_scl1
      fwd%pft(2,fwd%shot_pos)=fwd%pft(2,fwd%shot_pos)-par%src(jt)*src_scl2

      !! Damping function applied to the source wavefield
      if(use_taper) call taperings(fwd%pftm2,fwd%pftm1,fwd%pft)

      !! Obtain snapshots of the source wavefield for a single source
      if(par%snapsht .and. jsources == par%src_snap&
                     .and. mod(jt,jtsnap) == 0) then
        par%ntSnap=par%ntSnap+1
        call fwd%get_snaps_src(par%sgx(:,jsources))
      end if

      !! Extract the wavefield on the surface
      this%dsynt(jt,:)=this%dsynt(jt,:)+fwd%pft(1,:)
      this%dsynt(jt,:)=this%dsynt(jt,:)-fwd%pft(2,:)

    end do
  !! Get the timing
  end procedure

  module procedure nonlin_init
    !! Group some useful image limits
    x1 = 1-par%napL
    x2 = par%ngx+par%napR
    if(.not. allocated(this%dsynt)) then 
      allocate(this%dsynt(par%nt,x1-par%pad_:x2+par%pad_),stat=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'nonlin_init: could not allocate dsynt array!'
      end if
    end if
 
!    allocate(owt(par%ngx),stat=allocation_status)
!    if(allocation_status /= 0) then
!      call skull()
!      error stop 'nonlin_init: could not allocate owt array!'
!    end if

!    ! This is a tapering function with offset, employed by Ali (disabled now)
!    do jgx=1,par%ngx
!      owt(jgx) = -1.*abs(sx(par%trace_shot)-gx(jgx))
!    end do
!    owt = owt - minval(owt) + 1.
!    owt = sin(pi/2.*owt/maxval(owt))*sin(pi/2.*owt/maxval(owt)) 
  end procedure

end submodule
