submodule(read_mod) read_mod_implementation
  implicit none
contains

  module procedure read_parameters
    write(0,*) 'read_parameters: Reading parameters from command line'
    associate(pad_=>this%pad_,zpad=>this%zpad,vel_fracc=>this%vel_fracc,npts=>this%npts)
    associate(apertureL=>this%apertureL,apertureR=>this%apertureR,scalco=>this%scalco)
    associate(ntsData=>this%ntsData,nthreads=>this%nthreads,niter=>this%niter)
    associate(src_snap=>this%src_snap,cos2thc=>this%cos2thc)
    associate(alpha=>this%alpha,ispc=>this%ispc,override_sx_header=>this%override_sx_header)
    associate(nlagz=>this%nlagz,nlagx=>this%nlagx,incPSFz=>this%incPSFz,incPSFx=>this%incPSFx)
      !! Parameters read from the command line
      call from_param('vel_frac',vel_fracc,1.0)
      call from_param('wavelet_fn',wavelet_fn)
      call from_param('reflfile',string_aux,'none')
      if(string_aux(1:4) == 'none') then
        reflfile=' '
      else
        reflfile=string_aux
      end if
      call from_param('datafile',string_aux,'none')

      if(string_aux(1:4) == 'none') then
        datafile=' '
        call from_param('projpref',string_aux,'none')
        
        this%projpref=trim(adjustl(string_aux))
      else
        datafile=string_aux
        this%projpref="none"
      end if
      call from_param('npts',npts,0)
      call from_param('apertureL',apertureL,1000.0)
      call from_param('apertureR',apertureR,1000.0)
      call from_param('scalco_override',scalco,1)
      call from_param('frac_time',frac_time,1.)
    
      !! Number of threads if using openmp
      call from_param('nthreads',nthreads,1)
      !! Set the number of threads for OpenMP
!      call omp_set_num_threads(nthreads)
    
      !! Set process for a single shot 
      call from_param('one_src',char_aux,'n')
        this%one_src = (char_aux == 'y')
      call from_param('nts',ntsData,10)

      !! Verbose the source values?
      call from_param('verb_src',char_aux,'n')
        verb_src = (char_aux == 'y')

      !! Zero padding for the random boundary conditions
      call from_param('pad',pad_,20)
    
      !! Additional zero padding for top random boundary
      call from_param('zpad',zpad,30)
    
      !! Parameters for the energy imaging conditions
      call from_param('thetac',thetac,90)
      cos2thc=cos(2.*thetac*deg2rad)
      call from_param('alpha',alpha,0.0)

!      !! Set the use of ispc
!      call from_param('ispc',char_aux,'n')
!        ispc = (char_aux == 'y')

      call from_param('eic',char_aux,'n')
        this%eic = (char_aux == 'y')
    
      call from_param('override_sx_header',char_aux,'n')
        override_sx_header = (char_aux == 'y')
      !! ox_shift shifts all coordinates to make them divisible by sampling rate 
      call from_param('ox_shift',ox_shift,0.0)  
    
      !! Parameters for computing source snapshots
      call from_param('snapsht',char_aux,'n')
        this%snapsht = (char_aux == 'y')
      call from_param('src_snap',src_snap,1)
      if(this%snapsht) this%one_src=.true.

      !! Number of iterations for LWI
      call from_param('niter',niter,1)

      !! Number of lags for Hessian computation
      call from_param('nlagz',nlagz,0)
      call from_param('nlagx',nlagx,0)
      call from_param('incPSFz',incPSFz,1)
      call from_param('incPSFx',incPSFx,1)
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    write(0,*) 'read_parameters: Finishing reading parameters'
  end procedure

  module procedure read_history
    write(0,*) 'read_history: Reading parameters from history files'
    associate(o1=>this%o1,o2=>this%o2,d2=>this%d2,d3=>this%d3,unit1=>this%unit1,unit2=>this%unit2)
    associate(ntData=>this%ntData,nz=>this%nz,nx=>this%nx,dz=>this%dz,dx=>this%dx)
    associate(dtData=>this%dtData,dz=>this%dz,dx=>this%dx,t_end=>this%t_end)
    associate(ntracesData=>this%ntracesData)
    associate(first_cdp=>this%first_cdp,last_cdp=>this%last_cdp)
      !! Parameters read from the history file
      call from_history('n1',nz)
      call from_history('n2',nx)
      call from_history('o1',o1,0.00)
      call from_history('o2',o2,0.00)
      call from_history('d1',dz,0.00)
      call from_history('d2',d2,0.00)
      call from_history('d2',dx)
      call from_history('unit1',unit1,'')
      call from_history('unit2',unit2,'')

      !! Parameters read from auxiliary files

!      dummy=sep_get_header_format_tag(adjustr(datafile),tag_input)
!      if(dummy == 0) then
      if(datafile /= ' ') then
        call from_aux("datafile",'n1',ntData,0)
        ntData=nint(frac_time*ntData)
        call from_aux("datafile",'n2',ntracesData,0)
        call from_aux("datafile",'d1',dtData,0.00)
      end if
  
      !! Define some required parameters
      first_cdp = 1+floor((o2+ox_shift)/d2)
      last_cdp = first_cdp+(nx-1) 
      t_end = (ntData-1)*dtData
      z_start = o1
      z_end = o1+dz*(nz-1)
    end associate
    end associate
    end associate
    end associate
    end associate
    write(0,*) 'read_history: Finishing reading history files'
  end procedure

  module procedure read_headers
    write(0,*) 'Reading the input headers'
    associate(dx=>this%dx,dsx=>this%dsx,dgx=>this%dgx,dgxOdx=>this%dgxOdx)
    associate(ntracesData=>this%ntracesData,ntraces=>this%ntraces,scalco=>this%scalco)
      dummy=sep_get_header_format_tag("datafile",tag_input)
      if(dummy /= 0) then
        call skull()
        error stop 'read_headers: No dataset detected; aborting'
      else
        input_error=0
        write(0,*) 'Compute data with input parameters from data file'
        if(sep_get_number_keys("datafile",n_keys_data) /= 0) then
          write(0,*) 'Error, cannot read in number of keys from the hff (.H@@) file!'
          input_error=input_error+1
        end if
        if(sep_get_key_index("datafile","sx",sx_key) /= 0) then
          write(0,*) 'error! sx key index does not exist on datafile file!'
          input_error=input_error+1
        end if
        if(sep_get_key_index("datafile","gx",gx_key) /= 0) then
          write(0,*) 'error! gx key index does not exist on datafile file!'
          input_error=input_error+1
        end if
        if(sep_get_key_index("datafile","sy",sy_key) /= 0) then
          write(0,*) 'error! sy key index does not exist on datafile file!'
          input_error=input_error+1
        end if
        if(sep_get_key_index("datafile","gy",gy_key) /= 0) then
          write(0,*) 'error! gy key index does not exist on datafile file!'
          input_error=input_error+1
        end if
        if(sep_get_key_index("datafile","scalco",scalco_key) /= 0) then
          write(0,*) 'error! scalco key index does not exist on datafile file!'
          input_error=input_error+1
        end if
        allocate(header_in(n_keys_data,ntracesData))
        header_in = 0.
        if(sep_get_val_headers("datafile",1,ntracesData,header_in) /= 0) then
          call skull()
          error stop 'could not read from input data headers!'
        end if

        !! Reading input geometry from the keys
        if(scalco == 0) scalco = header_in(scalco_key,1)
      
        dsx = header_in(sx_key,2)-header_in(sx_key,1)
        dgx = header_in(gx_key,2)-header_in(gx_key,1)

        !! If header values are integer, they are read below
        if(dgx==0.) then
          allocate(headerI_in(n_keys_data,ntracesData))
          headerI_in = 0.
          if(sep_get_val_headers("datafile",1,ntracesData,headerI_in) /= 0) then
            call skull()
            error stop 'could not read from input data headers!'
          end if

          header_in=1.*headerI_in  
          deallocate(headerI_in)

          !! Reading input geometry from the keys
          if(scalco == 0) scalco = header_in(scalco_key,1)
        
          dsx = header_in(sx_key,2)-header_in(sx_key,1)
          dgx = header_in(gx_key,2)-header_in(gx_key,1)
        end if

        scalcof = scalco
        if(scalco < 0) then
          scalcof = -1./scalcof
        end if
        dgx = dgx*scalcof
        dgxOdx = ceiling(dgx/dx)
        if(dgxOdx == 1) then
          ntraces = ntracesData
        else
          ntraces = ntracesData*dgxOdx
        end if
        allocate(sx(ntraces),gx(ntraces),hoff(ntraces))
  
        do jtraces = 1,ntraces
          jfx = 1+(jtraces-1)/dgxOdx
          jx = jfx
          jfx = jfx-jx
          if(jx >= 1 .and. jx < ntracesData) then
            hsx_cur  = header_in(sx_key,jx)*scalcof
            hsx_next = header_in(sx_key,jx+1)*scalcof
            hgx_cur  = header_in(gx_key,jx)*scalcof
            hgx_next = header_in(gx_key,jx+1)*scalcof
            sx(jtraces) = hsx_cur
            if(abs(hgx_next-hgx_cur) == dgx) then
              gx(jtraces) = ((1.0-jfx)*hgx_cur+jfx*hgx_next)
            elseif(jfx == 0.0) then
              gx(jtraces) = hgx_cur
              gx(jtraces+1) = hgx_cur+dx
            end if
          elseif(jx == ntracesData) then
            sx(jtraces)   = header_in(sx_key,jx)*scalcof
            gx(jtraces)   = header_in(gx_key,jx)*scalcof
          end if
        end do
        sx = sx+ox_shift
        gx = gx+ox_shift
        deallocate(header_in)
      end if
  
      !! Calculate offsets withing the data geometry
      hoff = 0
      do jtraces = 1,ntraces
        hoff(jtraces) = abs(gx(jtraces)-sx(jtraces))
      end do
      hmin = minval(hoff)
      this%src_pad = floor(hmin/dx)
    end associate
    end associate
  end procedure

  module procedure read_velocity
    write(0,*) 'read_velocity: Begin reading velocities...'
    associate(first_cdp=>this%first_cdp,last_cdp=>this%last_cdp,nz=>this%nz)
      allocate(vinput(nz,first_cdp:last_cdp),&
               this%V_int(nz,first_cdp:last_cdp),&
               stat=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'read_velocity: could not allocate velocity arrays!'
      end if
      call sep_read(vinput)
      call assert(all(vinput > 0),&
         "read_velocity: zero or negative velocity values in the input file!")
      this%V_int=this%vel_fracc*vinput(:,first_cdp:last_cdp)
      deallocate(vinput)
      write(0,*) 'read_velocity: Finish reading velocities.'
      
      this%Vcte = sum(this%V_int(:,:))/(this%nx*this%nz)
      this%Vmin = minval(this%V_int)
      this%Vmax = maxval(this%V_int)
    end associate
  end procedure

  module procedure read_reflectivity
    write(0,*) 'read_reflectivity: Begin reading reflectivity...'
    associate(first_cdp=>this%first_cdp,last_cdp=>this%last_cdp,nz=>this%nz)
      allocate(vinput(nz,first_cdp:last_cdp),&
               this%refl(nz,first_cdp:last_cdp),&
               stat=allocation_status)
      this%refl=0.
      if(allocation_status /= 0) then
        call skull()
        error stop 'read_reflectivity: could not allocate reflectivity array!'
      end if
      if(reflfile/= ' ') call sep_read(vinput,reflfile)
      this%refl = vinput
      deallocate(vinput)
      if(reflfile/=' ') call assert(norm2(this%refl)/=0.,&
           "read_reflectivity: reflectivity values are zero!")
      write(0,*) 'read_reflectivity: Finish reading reflectivity.'
    end associate
  end procedure

  module procedure courant_stuff
    !! Get the Courant number; refine time-step and multiply time samples
    associate(dtData=>this%dtData,dt=>this%dt,dz=>this%dz,dx=>this%dx)
    associate(cfl=>this%cfl,t_end=>this%t_end,Vmax=>this%Vmax,nt=>this%nt)
      cfl = Vmax*dtData/min(dx,dz)
      k = ceiling(cfl/0.7)
      tIntFact = k
      dt = dtData/tIntFact
      nt = 1+nint(t_end/dt)
    end associate
    end associate
  end procedure

  module procedure read_src
    associate(dtData=>this%dtData,dt=>this%dt,dz=>this%dz,dx=>this%dx)
    associate(ntData=>this%ntData,nt=>this%nt,ntsData=>this%ntsData)
    associate(ngx=>this%ngx,jtStep=>this%jtStep)
    associate(x1=>1-this%pad_-this%napL,x2=>ngx+this%pad_+this%napR)
      if(.not. allocated(srcData)) then
        allocate(srcData(ntData),   &
                 rcvData(ntData),   &
                 this%src(nt),      &
                 this%rcv(nt,x1:x2),&
                 STAT=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'read_src: Could not allocate source arrays!'
        end if
      end if
      srcData = 0
      this%src = 0
      if(present(wavelet_fn_add)) then
        call sep_read(srcData,wavelet_fn_add)
      else
        call sep_read(srcData,wavelet_fn)
      end if
      if(verb_src) then
        write(0,*) 'Report source values for debuggin purposes'
        write(0,*) 'srcData ',srcData(1:ntsData)
      end if
      if(.not. present(noderiv)) srcData=srcData/dtData**2
      call interp(ntData,tIntFact,nt,srcData,this%src(:),.false.)
      this%nts = tIntFact*ntsData
      this%src(this%nts+1:nt) = 0
!      this%src(:)=this%src(:)/this%dt**2
      fNyq = 1/(2.0*dtData)
      call from_param('fc',fc,fNyq)
      !fc = min(fc,0.5*Vmin*SQRT(1.0/(dx*dx)+1.0/(dz*dz)))
      fc = max(fc,0.5*this%Vmin/max(dx,dz))
      jtStep = floor(1/(2*dt*fc))
      if(jtStep < 1) jtStep = 1 
    end associate
    end associate
    end associate
    end associate
  end procedure

  module procedure par_geometry
    write(0,*) ' '
    write(0,*) 'par_geometry: Compute data with input parameters from par file'
    associate(pad_=>this%pad_,zpad=>this%zpad,src_pad=>this%src_pad)
    associate(nsources=>this%nsources,ntData=>this%ntData,ntraces=>this%ntraces)
    associate(ngx=>this%ngx,nz=>this%nz,nkz=>this%nkz,nkgx=>this%nkgx,nbox=>this%nbox)
    associate(dx=>this%dx,dsx=>this%dsx,dgx=>this%dgx,dtData=>this%dtData,dgxOdx=>this%dgxOdx)
    associate(napL=>this%napL,napR=>this%napR)
    associate(t_end=>this%t_end,ntData=>ntData)
    associate(trace_shot=>this%trace_shot,ntraces=>this%ntraces,nsources=>this%nsources)
    associate(first_source=>this%first_source,last_source=>this%last_source)
    associate(one_src=>this%one_src,src_snap=>this%src_snap)
    associate(apertureL=>this%apertureL,apertureR=>this%apertureR)
      call from_param('nsources',nsources)
      call from_param('dsx',dsx)
      call from_param('osx',osx)
      call from_param('ngx',ngx)
      call from_param('dgx',dgx)
      call from_param('ogx',ogx)
      call from_param('ntData',ntData)
      call from_param('dtData',dtData)
  
      dgxOdx = ceiling(dgx/dx)
  
      if(dgxOdx /= 1) then
        ngx=ngx*dgxOdx
      end if
  
      ntraces = nsources*ngx
      allocate(sx(ntraces),gx(ntraces),hoff(ntraces))
      jtraces=0
      do jsources=1,nsources
        do jgx=1,ngx
          jtraces=jtraces+1
          sx(jtraces)=(osx+(jsources-1))*dsx
          gx(jtraces)=(ogx+(jgx-1))*dgx
        end do 
      end do 

      sx = sx+ox_shift
      gx = gx+ox_shift
    
      !! Calculate offsets withing the data geometry
      hoff = 0
      do jtraces = 1,ntraces
        hoff(jtraces) = abs(gx(jtraces)-sx(jtraces))
      end do
      hmin = minval(hoff)
      this%src_pad = floor(hmin/dx)
      t_end = (ntData-1)*dtData

      allocate(this%sgx(1-napL:ngx+napR+src_pad,nsources),&
               sgx_max(nsources),&
               source_index(nsources,2)) 
      write(0,*) 'Prepare geometric arrays for sources'
      this%sgx = UNDEFINED 
      k = 1
      l = 1
      ngx_p = 1
      do jtraces = 1,ntraces
        if(jtraces < ntraces) then
          if(sx(jtraces) == sx(jtraces+1)) then
            this%sgx(k,l) = gx(jtraces)
            k = k+1
          else
            if(k > ngx_p) then
              ngx = k
              ngx_p = k
            end if
            source_index(l,1) = k
            source_index(l,2) = jtraces
            this%sgx(k,l) = gx(jtraces)
            k = 1
            l = l+1
          end if
        else
          this%sgx(k,l) = gx(jtraces)
          source_index(l,1) = k-1
          source_index(l,2) = jtraces
      
          if(k < ngx) then
            k = k+1
          else
            if(k > ngx_p) then
              ngx = k
              ngx_p = k
            end if
            k = 1
            l = l+1
          end if
        end if
      end do
      sgx_max = maxval(this%sgx(:,:),dim=1)
      do jgx = 1,src_pad+napR
        this%sgx(ngx+jgx,:) = sgx_max(:)+jgx*dx
      end do
      do jgx = 1-napL,0
        this%sgx(jgx,:) = this%sgx(1,:)-(1-jgx)*dx
      end do
      !! Set the first and last source values      
      if(one_src) then
        !! Fix one source if snapshots of the wavefield are wanted
        first_source = src_snap
        last_source = src_snap
        trace_shot = 1+(src_snap-1)*ngx
      else
        first_source = 1
        last_source = nsources
        trace_shot = 1
      end if
      napL = floor(apertureL/dx)
      napR = floor(apertureR/dx)
      nbox = ngx+napL+napR
      nxpad = nbox+pad_+pad_
      nkgx = nxpad+src_pad
      nkz = nz+pad_+pad_+zpad
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
  end procedure

  module procedure data_geometry
    associate(pad_=>this%pad_,zpad=>this%zpad,src_pad=>this%src_pad)
    associate(ngx=>this%ngx,nz=>this%nz,nkz=>this%nkz,nkgx=>this%nkgx,nbox=>this%nbox)
    associate(trace_shot=>this%trace_shot,ntraces=>this%ntraces,nsources=>this%nsources)
    associate(first_source=>this%first_source,last_source=>this%last_source)
    associate(dx=>this%dx,napL=>this%napL,napR=>this%napR)
    associate(one_src=>this%one_src,src_snap=>this%src_snap)
    associate(apertureL=>this%apertureL,apertureR=>this%apertureR)
      !! Obtain geometry of the shot gathers ***
      k = 1
      l = 1
      ngx_p = 1
      do jtraces = 1,ntraces
        if(jtraces < ntraces) then
          if(sx(jtraces) == sx(jtraces+1)) then
            k = k+1
          else
            if(k > ngx_p) then
              ngx = k
              ngx_p = k
            end if
            k = 1
            l = l+1
          end if
          nsources = l
        elseif(sx(1) > sx(ntraces)) then
          if(k < ngx) then
            k = k+1
          else
            if(k > ngx_p) then
              ngx = k
              ngx_p = k
            end if
            k = 1
            l = l+1
          end if
        end if
      end do
      if(nsources == 1) ngx = k 
      napL = floor(apertureL/dx)
      napR = floor(apertureR/dx)
      nbox = ngx+napL+napR
      nxpad = nbox+pad_+pad_
      nkgx = nxpad+src_pad
      nkz = nz+pad_+pad_+zpad
      allocate(this%sgx(1-napL:ngx+napR+src_pad,nsources),&
               sgx_max(nsources),&
               source_index(nsources,2)) 
      write(0,*) 'Prepare geometric arrays for sources'
      this%sgx = UNDEFINED 
      k = 1
      l = 1
      ngx_p = 1
      do jtraces = 1,ntraces
        if(jtraces < ntraces) then
          if(sx(jtraces) == sx(jtraces+1)) then
            this%sgx(k,l) = gx(jtraces)
            k = k+1
          else
            if(k > ngx_p) then
              ngx = k
              ngx_p = k
            end if
            source_index(l,1) = k
            source_index(l,2) = jtraces
            this%sgx(k,l) = gx(jtraces)
            k = 1
            l = l+1
          end if
        else
          this%sgx(k,l) = gx(jtraces)
          source_index(l,1) = k-1
          source_index(l,2) = jtraces
      
          if(k < ngx) then
            k = k+1
          else
            if(k > ngx_p) then
              ngx = k
              ngx_p = k
            end if
            k = 1
            l = l+1
          end if
        end if
      end do
      sgx_max = maxval(this%sgx(:,:),dim=1)
      do jgx = 1,src_pad+napR
        this%sgx(ngx+jgx,:) = sgx_max(:)+jgx*dx
      end do
      do jgx = 1-napL,0
        this%sgx(jgx,:) = this%sgx(1,:)-(1-jgx)*dx
      end do

      !! Set the first and last source values      
      first_source = 1
      last_source = nsources
      trace_shot = 1
      if(one_src) then
        !! Fix one source if snapshots of the wavefield are wanted
        first_source = src_snap
        last_source = src_snap
        trace_shot = 1+(src_snap-1)*ngx
      end if
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    end associate
    write(0,*) 'End geometric arrays for sources'
  end procedure

  module procedure read_shot
    k = 1
    associate(first_geof=>this%first_geof,last_geof=>this%last_geof,jgeof=>this%jgeof,dgxOdx=>this%dgxOdx)
    associate(rcv=>this%rcv,ntData=>this%ntData,nt=>this%nt,ngx=>this%ngx)
      rcv=0.
      jgeof = ceiling(first_geof/dgxOdx)-1 
      do jx = first_geof,last_geof,dgxOdx
        jgeof = jgeof+1
        previous_trace = jgeof-1
  
        !! Reads data traces
        if(sseek_block("datafile",previous_trace,4*ntData,0) < 0) &
        call seperr('read_shot: Have a trouble seeking trace')
        if(sreed("datafile",rcvData,4*ntData) /= 4*ntData) &
        call seperr('read_shot: Have a trouble reading trace')

        !! Interpolate time samples to satisfy Courant condition
        call interp(ntData,tIntFact,nt,rcvData,rcv(:,k),.false.)
        k = k+1
      end do
    end associate
    end associate
  end procedure

  module procedure verbose_par
    write(0,*) '***************************'
    write(0,*) '* Verbose some parameters *'
    write(0,*) '***************************'
    write(0,*) 'nz ',this%nz
    write(0,*) 'ntData ',this%ntData
    write(0,*) 'nt ',this%nt
    write(0,*) 'ntsData ',this%ntsData
    write(0,*) 'nts ',this%nts
    write(0,*) 'nx ',this%nx
    write(0,*) 'ngx ',this%ngx
    write(0,*) 'nkz ',this%nkz
    write(0,*) 'nkgx ',this%nkgx
    write(0,*) 'nbox ',this%nbox
    write(0,*) 'ntraces ',this%ntraces
    write(0,*) 'nsources ',this%nsources
    write(0,*) 'Points triang filt ',this%npts
    write(0,*) 'first_cdp ',this%first_cdp
    write(0,*) 'last_cdp ',this%last_cdp
    write(0,*) 'pad_',this%pad_
    write(0,*) 'src_pad',this%src_pad
    write(0,*) 'Global av. vel. ',this%Vcte
    write(0,*) 'Vel. min. ',this%Vmin
    write(0,*) 'Vel. max. ',this%Vmax
    write(0,*) 'Vel. fracc. ',this%vel_fracc
    write(0,*) 'Mult. time step ',this%jtStep
    write(0,*) 'Max. freq. in data ',fc
    write(0,*) 'dgx ',this%dgx
    write(0,*) 'dx ',this%dx
    write(0,*) 'dtData ',this%dtData
    write(0,*) 'dt ',this%dt
    write(0,*) 'dz ',this%dz
    write(0,*) 'dgxOdx ',this%dgxOdx
    write(0,*) 'Courant number ',this%cfl
    write(0,*) 'Aperture left',this%apertureL
    write(0,*) 'Aperture right',this%apertureR
    write(0,*) 'napL',this%napL
    write(0,*) 'napR',this%napR
    write(0,*) 'scalco ',this%scalco
    write(0,*) 'nthreads ',this%nthreads
    write(0,*) 'snapsht ',this%snapsht
    write(0,*) 'src_snap ',this%src_snap
    write(0,*) 'Use ispc? ',this%ispc
    write(0,*) 'No. iters (LWI only)? ',this%niter
    write(0,*) 'Current image ',this_image()
    write(0,*) 'Num images ',num_images()
    write(0,*) ' '

  end procedure
end submodule
