module rtm_mod
  !! This module implements conventional correlation imaging conditions and
  !! energy imaging conditions, and build the migrated images
  use declare_variables_mod, only : use_taper
  use read_mod, only : read_
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
!  use omp_lib
  use base_trismt_mod
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  private
  public :: rtm_image

  integer :: dummy, allocation_status
  integer :: k, jt, jtCoarse, jx, jz, jgx, zshift, x1, x2, x3, x4, z1, z2, z3, z4
  integer :: nt, nx, ngx, nz, first_cdp, last_cdp
  integer, parameter :: jtsnap=10 
  real    :: fx, jfx, tinyden, tinyden0, src_scl1, src_scl2
  real, allocatable, dimension(:)   :: owt
  real, allocatable, dimension(:,:) :: rcv_scl
  real, parameter :: tiny_=1e-3
  real, parameter :: pi = 4.*atan(1.) ! The Order of Chaos (Darren Aronofsky, 1998)
  logical, parameter :: T=.true., F=.false.
  
  type rtm_image
    !! This class has and numerators and denominators to create the migrated images 
    real, dimension(:,:), allocatable :: ICorr, ITomo, IRefl
    real, dimension(:,:), allocatable :: num, num0, num90, den, den0
  contains
    procedure :: rtm
    procedure :: rtm_init
    procedure :: update_imag
    procedure :: xcorrIC
    procedure :: energyIC
    procedure :: build_images
  end type

  interface
    pure module subroutine rtm(this,par,fwd,bwd,jsources)
      !! I apply the rtm procedure here
      implicit none
      class(rtm_image), intent(inout) :: this
      type(read_), intent(inout) :: par
      type(src_prop_fwd),     intent(inout) :: fwd 
      type(src_rcv_prop_bwd), intent(inout) :: bwd
      integer, intent(in) :: jsources
    end subroutine

    pure module subroutine rtm_init(this,par,fwd,bwd,sx,gx)
      !! Here the scalar variables are initialized
      implicit none
      class(rtm_image), intent(inout) :: this
      type(src_prop_fwd), intent(inout) :: fwd 
      type(src_rcv_prop_bwd), intent(inout) :: bwd 
      type(read_), intent(in) :: par
      real, dimension(:), intent(in) :: sx, gx
    end subroutine
 
    pure module subroutine update_imag(this,par)
      !! Simple procedure for setting arrays to zero every source iteration
      implicit none
      class(rtm_image), intent(inout) :: this
      type(read_), intent(in) :: par
    end subroutine

    pure module subroutine xcorrIC(this,par,bwd)
      !! Performs cross-correlation imaging condition (Claerbout, 1971)
      implicit none
      class(rtm_image), intent(inout) :: this
      type(src_rcv_prop_bwd), intent(in) :: bwd 
      type(read_), intent(in) :: par
    end subroutine

    pure module subroutine energyIC(this,par,bwd)
      !! Performs energy imaging conditions (e.g. Rocha et al., 2015),
      !! creating numerators and deniminators for deconvolution approach
      implicit none
      class(rtm_image), intent(inout) :: this
      type(src_rcv_prop_bwd), intent(inout) :: bwd 
      type(read_), intent(in) :: par
    end subroutine

    pure module subroutine build_images(this,par,sgx,conv)
      !! Constructs the images dividing corresponding numerators by denominators 
      implicit none
      class(rtm_image), intent(inout) :: this
      type(read_), intent(in) :: par
      real, intent(in) :: sgx(:)
      logical, intent(in) :: conv
    end subroutine
  end interface
end module
