submodule(taper_mod) taper_mod_implementation
  implicit none

contains
  module procedure taper_init
    allocate(par,source=par_in,stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'taper_init: Could not allocate type bound par!'
    end if
    associate(pad_=>par%pad_)
      allocate(zxtaper(pad_),&
               ctaper(pad_,pad_))
      do jj = 1,pad_
        zxtaper(jj) = 0.99+0.01*cos(pi*(pad_-jj)/pad_)
      end do
      do jx = 1,pad_
        do jz = 1,pad_
          dist = sqrt(1.001*(pad_-jx)*(pad_-jx)/pad_/pad_+&
                      0.999999*(pad_-jz)*(pad_-jz)/pad_/pad_)
          ctaper(jz,jx) = 0.99+0.01*cos(pi*dist)
        end do
      end do
      use_taper=.true.
    end associate
  end procedure

  module procedure taperings
    ! Corner
    associate(pad_=>par%pad_,nkz=>par%nkz,nkgx=>par%nkgx)
      do jx = 1,pad_
        do jz = 1,pad_
          prev(jz,jx) =              prev(jz,jx)*ctaper(jz,jx)
          prev(jz,nkgx-jx+1) =       prev(jz,nkgx-jx+1)*ctaper(jz,jx)
          prev(nkz-jz+1,nkgx-jx+1) = prev(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
          prev(nkz-jz+1,jx) =        prev(nkz-jz+1,jx)*ctaper(jz,jx)
  
          cur(jz,jx) =              cur(jz,jx)*ctaper(jz,jx)
          cur(jz,nkgx-jx+1) =       cur(jz,nkgx-jx+1)*ctaper(jz,jx)
          cur(nkz-jz+1,nkgx-jx+1) = cur(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
          cur(nkz-jz+1,jx) =        cur(nkz-jz+1,jx)*ctaper(jz,jx)
  
          next(jz,jx) =              next(jz,jx)*ctaper(jz,jx)
          next(jz,nkgx-jx+1) =       next(jz,nkgx-jx+1)*ctaper(jz,jx)
          next(nkz-jz+1,nkgx-jx+1) = next(nkz-jz+1,nkgx-jx+1)*ctaper(jz,jx)
          next(nkz-jz+1,jx) =        next(nkz-jz+1,jx)*ctaper(jz,jx)
        end do
      end do
  
      do jz = pad_+1,nkz-pad_
        prev(jz,1:pad_) =           prev(jz,1:pad_)*zxtaper(1:pad_:+1)
        prev(jz,nkgx-pad_+1:nkgx) = prev(jz,nkgx-pad_+1:nkgx)*zxtaper(pad_:1:-1)
  
        cur(jz,1:pad_) =           cur(jz,1:pad_)*zxtaper(1:pad_:+1)
        cur(jz,nkgx-pad_+1:nkgx) = cur(jz,nkgx-pad_+1:nkgx)*zxtaper(pad_:1:-1)
  
        next(jz,1:pad_) =           next(jz,1:pad_)*zxtaper(1:pad_:+1)
        next(jz,nkgx-pad_+1:nkgx) = next(jz,nkgx-pad_+1:nkgx)*zxtaper(pad_:1:-1)
      end do
  
      do jx = pad_+1,nkgx-pad_
        prev(1:pad_,jx) =         prev(1:pad_,jx)*zxtaper(1:pad_:+1)
        prev(nkz-pad_+1:nkz,jx) = prev(nkz-pad_+1:nkz,jx)*zxtaper(pad_:1:-1)
  
        cur(1:pad_,jx) =         cur(1:pad_,jx)*zxtaper(1:pad_:+1)
        cur(nkz-pad_+1:nkz,jx) = cur(nkz-pad_+1:nkz,jx)*zxtaper(pad_:1:-1)
  
        next(1:pad_,jx) =         next(1:pad_,jx)*zxtaper(1:pad_:+1)
        next(nkz-pad_+1:nkz,jx) = next(nkz-pad_+1:nkz,jx)*zxtaper(pad_:1:-1)
      end do
    end associate
  end procedure

end submodule
