submodule(vshot_mod) vshot_mod_implementation
  implicit none
contains

  module procedure vshot_init
    associate(pad_=>par%pad_,zpad=>par%zpad,dz=>par%dz,dx=>par%dx)
      allocate(distZU(1-pad_-zpad:0), &
               distZD(1:pad_),        &
               distXL(1-pad_:0),      &
               distXR(1:pad_),        & 
               distU(1-pad_-zpad:0),  &
               dist(1:pad_+zpad),     &
               stat=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'vshot_init: Could not allocate arrays!'
      end if

      shftx = pad_
      shftz = pad_+zpad
      distZU = 0
      distZD = 0
      distXL = 0 
      distXR = 0
  
      do jz =1-pad_-zpad,0,1
        distZU(jz) = fact*dz*(-jz)
      end do 
      do jz =1,pad_
        distZD(jz) = fact*dz*jz
      end do 
      do jx =1-pad_,0,1
        distXL(jx) = fact*dx*(1-jx)
      end do 
      do jx =1,pad_
        distXR(jx) = fact*dx*jx
      end do 

      distMax = pad_ 
      distMaxU = (pad_+zpad) 
      do jz =1-pad_-zpad,0,1
        distU(jz) = -1.0*jz
      end do 
      do jz =1,pad_+zpad
        dist(jz) = 1.0*(jz-1)
      end do 
    end associate
  end procedure
  
  module procedure vshot
!    associate(first_cdp=>par%first_cdp,last_cdp=>par%last_cdp)
!    associate(pad_=>par%pad_,zpad=>par%zpad,src_pad=>par%src_pad)
!    associate(nz=>par%nz,nbox=>par%nbox,dz=>par%dz,dx=>par%dx)
    associate(V=>par%V_int)
      if(minval(par%V_int) < 0.) then
        V_min = minval(par%V_int)
      else
        V_min = 0.
      end if
      V_max = maxval(par%V_int)

      V_shot  = 0
      fact = par%dx/par%dz
      if(fact < 1.0) then
        fact = 1.25/fact
      else
        fact = 1.25*fact
      end if
      
      call date_and_time(values=values)
      call random_seed(size=k)
      allocate(seed(1:k))
      seed(:) = values(8)
      call random_seed(put = seed)
      deallocate(seed)

      damp1 = 1
      damp2 = 1
      frac = 1.0 
   
      do jgx = 1-par%pad_,par%nbox+par%pad_+par%src_pad
        if(jgx >= 1 .and. jgx <= par%nbox+par%src_pad) then
          jx = 1+floor(sgx(jgx)/par%dx)
          if(jx >= par%first_cdp .and. jx <= par%last_cdp) then
            V_shot (1+shftz:par%nz+shftz,jgx+shftx) = V(1:par%nz,jx)
            do jz = 1-par%zpad,0
              V_shot (jz+shftz,jgx+shftx) = V(1,jx)
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do 
                rand1 = 0
                damp1 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  call random_number(rand1)
                  rand1 = 2*(rand1-0.5) 
                  damp1 = sin(0.5*pi*damp1)
!                else if(anis) then
!                  damp1 = 1+cos(0.5*pi*damp1)
                end if
                Vtest = V(1,jx)*damp1+rand1*distZU(jz)
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                damp1 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  call random_number(rand1)
                  rand1 = 2*(rand1-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                end if
                Vtest = V(par%nz,jx)*damp1+rand1*distZD(jz-par%nz)
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          elseif(jx < par%first_cdp) then
            V_shot(1+shftz:par%nz+shftz,jgx+shftx) = V(1:par%nz,par%first_cdp)
            do jz = 1-par%zpad,0
              V_shot(jz+shftz,jgx+shftx) = V(1,par%first_cdp)
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                damp1 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  call random_number(rand1)
                  rand1 = 2*(rand1-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                end if
                Vtest = V(1,par%first_cdp)*damp1+rand1*distZU(jz)
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                damp1 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  call random_number(rand1)
                  rand1 = 2*(rand1-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                end if
                Vtest = (V(par%nz,par%first_cdp)*damp1+rand1*distZD(jz-par%nz))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          elseif(jx > par%last_cdp) then
            V_shot (1+shftz:par%nz+shftz,jgx+shftx) = V(1:par%nz,par%last_cdp)
            do jz = 1-par%zpad,0
              V_shot (jz+shftz,jgx+shftx) = V(1,par%last_cdp)
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                damp1 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  call random_number(rand1)
                  rand1 = 2*(rand1-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                end if
                Vtest = V(1,par%last_cdp)*damp1+rand1*distZU(jz)
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                damp1 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  call random_number(rand1)
                  rand1 = 2*(rand1-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                end if
                Vtest = V(par%nz,par%last_cdp)*damp1+rand1*distZD(jz-par%nz)
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          end if
        elseif(jgx < 1) then
          jx = 1+floor(sgx(1)/par%dx)
          if(jx >= par%first_cdp .and. jx <= par%last_cdp) then
            do jz = -par%zpad+1,par%nz
              do
                rand2 = 0
                damp2 = 1.
                if(rand) then
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand2)
                  rand2 = 2*(rand2-0.5) 
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(jz < 1) then
                  Vtest = V(1,jx)*damp2+rand2*distXL(jgx)
                else
                  Vtest = V(jz,jx)*damp2+rand2*distXL(jgx)
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  Vtest = (V(1,jx)*damp1*damp2+rand1*distZU(jz)+rand2*distXL(jgx))
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(par%nz,jx)*damp1*damp2+rand1*distZD(jz-par%nz)+rand2*distXL(jgx))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                   V_shot(jz+shftz,jgx+shftx) = Vtest
                   exit
                end if
              end do
            end do
          elseif(jx < par%first_cdp) then
            do jz = -par%zpad+1,par%nz
              do
                rand2 = 0
                damp2 = 1.
                if(rand) then
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand2)  !! ALERT ***
                  rand2 = 2*(rand2-0.5) 
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(jz < 1) then
                  Vtest = (V(1,par%first_cdp)*damp2+rand2*distXL(jgx))
                else
                  Vtest = (V(jz,par%first_cdp)*damp2+rand2*distXL(jgx))
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 =frac*(1-distU(jz)/distMaxU)
                  damp2 =frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(1,par%first_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXL(jgx))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(par%nz,par%first_cdp)*damp1*damp2+rand1*distZD(jz-par%nz)+rand2*distXL(jgx))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          elseif(jx > par%last_cdp) then
            do jz = -par%zpad+1,par%nz
              do
                rand2 = 0
                damp2 = 1.
                if(rand) then
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand2)
                  rand2 = 2*(rand2-0.5) 
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(jz < 1) then
                  Vtest = (V(1,par%last_cdp)*damp2+rand2*distXL(jgx))
                else
                  Vtest = (V(jz,par%last_cdp)*damp2+rand2*distXL(jgx))
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(1,par%last_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXL(jgx))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  damp2 = frac*(1-dist(1-jgx)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(par%nz,par%last_cdp)*damp1*damp2+rand1*distZD(jz-par%nz)+rand2*distXL(jgx))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          end if
        else
          jx = 1+floor(sgx(par%nbox+par%src_pad)/par%dx)
          if(jx >= par%first_cdp .and. jx <= par%last_cdp) then
            do jz = -par%zpad+1,par%nz
              do
                rand2 = 0
                damp2 = 1.
                if(rand) then
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand2)
                  rand2 = 2*(rand2-0.5) 
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(jz < 1) then
                  Vtest = (V(1,jx)*damp2+rand2*distXR(jgx-par%nbox-par%src_pad))
                else
                  Vtest = (V(jz,jx)*damp2+rand2*distXR(jgx-par%nbox-par%src_pad))
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(1,jx)*damp1*damp2+rand1*distZU(jz)+rand2*distXR(jgx-par%nbox-par%src_pad))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(par%nz,jx)*damp1*damp2+rand1*distZD(jz-par%nz)+rand2*distXR(jgx-par%nbox-par%src_pad))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          elseif(jx < par%first_cdp) then
            do jz = -par%zpad+1,par%nz
              do
                rand2 = 0
                damp2 = 1.
                if(rand) then
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand2)
                  rand2 = 2*(rand2-0.5) 
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(jz < 1) then
                  Vtest = (V(1,par%first_cdp)*damp2+rand2*distXR(jgx-par%nbox-par%src_pad))
                else
                  Vtest = (V(jz,par%first_cdp)*damp2+rand2*distXR(jgx-par%nbox-par%src_pad))
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(1,par%first_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXR(jgx-par%nbox-par%src_pad))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(par%nz,par%first_cdp)*damp1*damp2+rand1*distZD(jz-par%nz)+rand2*distXR(jgx-par%nbox-par%src_pad))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          elseif(jx > par%last_cdp) then
            do jz = -par%zpad+1,par%nz
              do
                rand2 = 0
                damp2 = 1.
                if(rand) then
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand2)
                  rand2 = 2*(rand2-0.5) 
                  damp2 = sin(0.5*pi*damp2)
                end if
                if(jz < 1) then
                  Vtest = (V(1,par%last_cdp)*damp2+rand2*distXR(jgx-par%nbox-par%src_pad))
                else
                  Vtest = (V(jz,par%last_cdp)*damp2+rand2*distXR(jgx-par%nbox-par%src_pad))
                end if
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = 0,1-par%pad_-par%zpad,-1
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-distU(jz)/distMaxU)
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(1,par%last_cdp)*damp1*damp2+rand1*distZU(jz)+rand2*distXR(jgx-par%nbox-par%src_pad))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
            do jz = par%nz+1,par%nz+par%pad_
              do
                rand1 = 0
                rand2 = 0
                damp1 = 1.
                damp2 = 1.
                if(rand) then
                  damp1 = frac*(1-dist(jz-par%nz)/distMax)
                  damp2 = frac*(1-dist(jgx-par%nbox-par%src_pad)/distMax)
                  call random_number(rand1)
                  call random_number(rand2)
                  rand1 = 2*(rand1-0.5) 
                  rand2 = 2*(rand2-0.5) 
                  damp1 = sin(0.5*pi*damp1)
                  damp2 = sin(0.5*pi*damp2)
                end if
                Vtest = (V(par%nz,par%last_cdp)*damp1*damp2+rand1*distZD(jz-par%nz)+rand2*distXR(jgx-par%nbox-par%src_pad))
                if(Vtest <= V_max .and. Vtest >= V_min) then
                  V_shot(jz+shftz,jgx+shftx) = Vtest
                  exit
                end if
              end do
            end do
          end if
        end if
      end do
    end associate
!    end associate
!    end associate
  end procedure

end submodule
