module wemva_mod
  !! This module implements wave-equation migration vel. analysis
  use sep
  use declare_variables_mod, only : use_taper
  use read_mod, only : read_
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
!  use omp_lib
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  private
  public :: wemva_foo

  integer :: dummy, allocation_status
  integer :: k, jt, jtCoarse, jx, jz, jgx, zshift, x1, x2, x3, x4, z1, z2
  integer :: nt, nx, ngx, nz, nkgx, nkz, first_cdp, last_cdp
  integer, parameter :: jtsnap=10 
  real :: fx, jfx, src_scl1, src_scl2
  real, allocatable, dimension(:,:) :: scat_scl, tmp
  real, allocatable, dimension(:,:) :: rcv_scl, rcv_tt
  real, allocatable, dimension(:)   :: owt
  real, parameter :: pi = 4.*atan(1.) ! The Order of Chaos (Darren Aronofsky, 1998)
  logical, parameter :: T=.true., F=.false., test=.false.
  type(src_prop_fwd), allocatable :: scat

  type wemva_foo
    !! This class has and WEMVA-related variables
    real, dimension(:,:), allocatable :: pertshot, resd, perturb, IPert
  contains
    procedure :: wemva
    procedure :: wemva_init
    procedure :: build_perturb
  end type

  interface
    pure module subroutine wemva(this,par,fwd,bwd,vel,ref,jsources,temp)
      !! I apply wemva modeling procedure here
      implicit none
      class(wemva_foo), intent(inout) :: this
      type(src_prop_fwd),     intent(inout) :: fwd 
      type(src_rcv_prop_bwd), intent(inout) :: bwd 
      type(read_), intent(inout) :: par
      type(vshot_foo), intent(inout) :: vel
      type(rshot_foo), intent(inout) :: ref
      integer, intent(in) :: jsources
      real, optional, intent(in) :: temp(:,:)
    end subroutine

    pure module subroutine wemva_init(this,par,fwd,sx,gx)
      !! Here the scalar variables are initialized
      implicit none
      class(wemva_foo), intent(inout) :: this
      class(src_prop_fwd), intent(inout) :: fwd
      type(read_), intent(in) :: par
      real, dimension(:), intent(in) :: sx, gx
    end subroutine

    pure module subroutine build_perturb(this,par,sgx)
      !! Constructs the perturbations onto the model grid
      implicit none
      class(wemva_foo), intent(inout) :: this
      real, intent(in) :: sgx(:)
      type(read_), intent(in) :: par
    end subroutine
  end interface
end module
