submodule(wemva_mod) wemva_mod_implementation
  implicit none
contains
  module procedure wemva
    !! Create velocity frames for scatter wavefield propagation
    call vel%vshot(par,scat%Vshot,par%sgx(:,jsources),T)

    !! Source and receiver scaling factors (using dipole sources)
    associate(z1=>par%zpad+par%pad_,x1=>par%pad_+par%napL+bwd%shot_pos)
      src_scl1=-(fwd%Vshot(1+z1,x1)*par%dt)**2
      src_scl2=-(fwd%Vshot(2+z1,x1)*par%dt)**2
    end associate
    if(.not. allocated(rcv_scl)) allocate(rcv_scl(par%nkz,par%nkgx))
    rcv_scl=-(fwd%Vshot(:,:)*par%dt)**2
    if(.not. allocated(scat_scl)) allocate(scat_scl(par%nkz,par%nkgx))
    scat_scl=-(fwd%Vshot(:,:)*par%dt)**2

!DELETE AFTER TEST
    if(present(temp)) this%pertshot=temp
    call sep_write(this%pertshot,'rrandom')

    write(0,*) 'Perform wemva modelling'
    !! Initialize/update wavefield arrays
    call  fwd%init_src_wavefields()
    call  bwd%init_src_wavefields()
    call scat%init_src_wavefields()
    this%perturb=0.
    par%ntSnap=0

    !! 1st stage: Forward propagation of source wavefield
    do jt=1,nt+2
      !! Forward propagation of source wavefield
      if(jt>1) call fwd%fwd_prop(F)  !1 

      !! Inject the source
      fwd%pft(1,bwd%shot_pos)=fwd%pft(1,bwd%shot_pos)+par%src(jt)*src_scl1
      fwd%pft(2,bwd%shot_pos)=fwd%pft(2,bwd%shot_pos)-par%src(jt)*src_scl2

      !! Obtain snapshots of the source wavefield for a single source
      if(par%snapsht .and. jsources == par%src_snap&
                     .and. mod(jt,jtsnap) == 0) then
        par%ntSnap=par%ntSnap+1
        call fwd%get_snaps_src(par%sgx(:,jsources))
      end if
    end do
    !! Prepare the source wavefield for back-propagation
    bwd%pft=fwd%pftm1 
    bwd%pftm1=fwd%pft

    !! 2nd stage: Backward propagation of receiver wavefield + scattering
    !! and xcorr with back-propagated source wavefield
    do jt=nt+2,1,-1
      !! Inject the data
      if(jt<=nt) then
        bwd%prt(1,:)=bwd%prt(1,:)+par%rcv(jt,:)
        bwd%prt(2,:)=bwd%prt(2,:)-par%rcv(jt,:)
      end if

      !! Backward propagation of scattered source and receiver wavefields
      call bwd%bwd_prop(fwd)   !2
      if(par%eic) call bwd%grad_rcv(fwd) 

      !! Re-inject the source after scaling
      bwd%pft(1,bwd%shot_pos)=bwd%pft(1,bwd%shot_pos)+par%src(jt)*src_scl1
      bwd%pft(2,bwd%shot_pos)=bwd%pft(2,bwd%shot_pos)-par%src(jt)*src_scl2

      !! Scatter the receiver wavefield upon the perturbation in the image  
      if(par%eic) then
        scat%pft=scat%pft+(bwd%prt*rcv_scl)*this%pertshot
      else
        scat%pft=scat%pft+(bwd%d2prt*rcv_scl)*this%pertshot
      end if

      !! Propagate the scattered receiver wavefield
      call scat%fwd_prop(T)   !3
      if(par%eic) call scat%grad()

      !! Damping function applied to the scattered receiver wavefield
      if(use_taper) call taperings(scat%pftm2,scat%pftm1,scat%pft)
      if(use_taper) call taperings(scat%dpftm2,scat%dpftm1,scat%dpft)
      if(use_taper) call taperings(scat%pftX,scat%pftZ,scat%d2pft)
      if(use_taper) call taperings(scat%d2pftm2,scat%d2pftm1,scat%d2pft)

      !! Apply energy imaging conditions: S*deltaR
      if(par%eic) then
        if(jt<=nt) this%perturb=this%perturb&
               +(bwd%pftX*scat%pftX+bwd%pftZ*scat%pftZ)*scat_scl&
               +par%cos2thc*(bwd%dpft*scat%dpft)*scat_scl/bwd%Vshotr**2
      else
        !! Apply imaging condition: S*deltaR
        if(jt<=nt) this%perturb=this%perturb+bwd%pft*(scat%pft*scat_scl)
      end if
!      !! Obtain snapshots of the receiver wavefield for a single source
!      if(jt<=nt .and. par%snapsht .and. jsources == par%src_snap&
!                     .and. mod(jt,jtsnap) == 0) then
!!        par%ntSnap=par%ntSnap+1
!!        call scat%get_snaps_scat(par%sgx(:,jsources))
!        call bwd%get_snaps_rcv(par%sgx(:,jsources))
!      end if
    end do

    !! Prepare the receiver wavefield for forward-propagation
    tmp=bwd%prt 
    bwd%prt=bwd%prtp1
    bwd%prtp1=tmp

    tmp=bwd%dprt 
    bwd%dprt=bwd%dprtp1
    bwd%dprtp1=tmp

    tmp=bwd%d2prt 
    bwd%d2prt=bwd%d2prtp1
    bwd%d2prtp1=tmp

    !! Re-initialize source wavefield arrays
    call fwd%init_src_wavefields()
    call scat%init_src_wavefields()
    bwd%pft=0.
    bwd%pftm1=0.

    !! 3rd stage: Forward propagation of source wavefield + scattering
    !! and xcorr with forward-propagated receiver wavefield
    do jt=1,par%nt
      !! Re-inject the data
      if(jt>2) then
        bwd%prt(1,:)=bwd%prt(1,:)+par%rcv(jt-2,:)
        bwd%prt(2,:)=bwd%prt(2,:)-par%rcv(jt-2,:)
      end if

      !! Forward propagation of source and receiver wavefields
      call bwd%bwd_prop(fwd)   !4
      if(par%eic) call bwd%grad_rcv(fwd) 

      !! Inject the source
      bwd%pft(1,bwd%shot_pos)=bwd%pft(1,bwd%shot_pos)+par%src(jt)*src_scl1
      bwd%pft(2,bwd%shot_pos)=bwd%pft(2,bwd%shot_pos)-par%src(jt)*src_scl2

      !! Damping function applied to the source wavefield
      if(use_taper) call taperings(bwd%pftm2,bwd%pftm1,bwd%pft)

      !! Scatter the source wavefield upon the reflectivity
      scat%pft=scat%pft+(bwd%pft*scat_scl)*this%pertshot

      !! Propagate the scattered  source wavefield
      call scat%fwd_prop(F)   !5
      if(par%eic) call scat%grad() 

      !! Damping function applied to the scattered source wavefield
      if(use_taper) call taperings(scat%pftm2,scat%pftm1,scat%pft)
      if(use_taper) call taperings(scat%dpftm2,scat%dpftm1,scat%dpft)
      if(use_taper) call taperings(scat%pftX,scat%pftZ,scat%d2pft)
      if(use_taper) call taperings(scat%d2pftm2,scat%d2pftm1,scat%d2pft)

      !! Obtain snapshots of the receiver wavefield for a single source
      if(par%snapsht .and. jsources == par%src_snap&
                     .and. mod(jt,jtsnap) == 0) then
        call bwd%get_snaps_rcv(par%sgx(:,jsources))
!        call scat%get_snaps_scat(par%sgx(:,jsources))
      end if

      if(par%eic) then
        !! Apply energy imaging conditions: R*deltaS
        this%perturb=this%perturb&
               +(bwd%prtX*scat%pftX+bwd%prtZ*scat%pftZ)*scat_scl&
               +par%cos2thc*(bwd%dprt*scat%dpft)*scat_scl/bwd%Vshotr**2
      else
        !! Apply imaging condition: R*deltaS
        this%perturb=this%perturb+(bwd%d2prt*rcv_scl)*scat%pft
      end if
    end do
  end procedure

  module procedure wemva_init
    !! Group some useful image limits
    x1 = 1-par%napL
    x2 = par%ngx+par%napR
    x3 = x1-par%pad_
    x4 = x2+par%pad_
    z1 = 1-par%zpad-par%pad_
    z2 = par%nz+par%pad_
    nt=par%nt
    nx=par%nx
    ngx=par%ngx
    nz=par%nz
    nkgx=par%nkgx
    nkz=par%nkz
    first_cdp=par%first_cdp
    last_cdp=par%last_cdp

    allocate(this%perturb(z1:z2,x3:x4), &
             this%IPert(nz,first_cdp:last_cdp), &
             tmp(nkz,nkgx),&
             stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'wemva_init: could not allocate arrays!'
    end if
    this%IPert=0

    !! Allocate arrays for scattered wavefield computations 
    allocate(this%pertshot,mold=fwd%Vshot, stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'born_init: could not perform mold allocation of pertshot!'
    end if

    allocate(scat,source=fwd,stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'born_init: could not assign parameter object "scat"'
    end if
    if(par%eic) call scat%grad_init()

    allocate(owt(par%ngx),stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'born_init: could not allocate owt array!'
    end if

    ! This is a tapering function with offset, employed by Ali (disabled now)
    do jgx=1,par%ngx
      owt(jgx) = -1.*abs(sx(par%trace_shot)-gx(jgx))
    end do
    owt = owt - minval(owt) + 1.
    owt = sin(pi/2.*owt/maxval(owt))*sin(pi/2.*owt/maxval(owt)) 
  end procedure

  module procedure build_perturb
    k=0
    !$omp parallel do default(shared) private(jgx,k,jx,fx)
    do jgx = x1,x2
      k=jgx-x1+1
      fx = nint(1.5001+(sgx(k)-0.0)/par%dx)
      jx = fx
      fx = fx-jx 
      if(jx >= first_cdp .and. jx < last_cdp) then
        this%IPert(:,jx)=this%IPert(:,jx)+(1-fx)*this%perturb(1:par%nz,jgx)&
                                         +   fx *this%perturb(1:par%nz,jgx+1)
        !! Test for NaNs
        call assert(.not. any(isnan(this%IPert)),"build_perturb: NaNs detected in IPert")
      end if
    end do
    !$omp end parallel do
  end procedure

end submodule
