module write_mod
  !! This subroutine writes the migration output: binary and SEP files
  use sep
  use declare_variables_mod, only : sx, gx
  use skull_mod
  use src_rcv_interp_mod
  use read_mod,   only : read_
  use born_mod,   only : born_foo
  use nonlin_mod, only : nonlin_foo
  use rtm_mod,    only : rtm_image 
  use wemva_mod,  only : wemva_foo
  use assertions
  implicit none
  private
  public :: write_

  integer :: n_nonzero_samples, shot_num, allocation_status, dummy
  integer :: output_error, jsrc, previous_trace, jgx
  integer, external :: sseek_block, srite
  integer, dimension(:,:), allocatable :: header_out
  integer, external :: sep_put_val_headers
  integer, external :: sep_put_number_keys, sep_put_key
  integer, external :: sep_put_data_axis_par, sep_put_header_axis_par
  real    :: amp_rms, amp_mean, min_amplitude, max_amplitude, scalcof
  real, allocatable :: d(:)
  character(len=4) :: data_fn="data"
  character(len=:), allocatable :: shot_fn(:)
  character(len=:), allocatable :: temp_string, out_,tag_input
  
  type write_
  !! This class contains the type-bound procedures for writing the output from
  !! rtm and born modelling
  contains
    procedure, nopass :: put_headers_data
    procedure, nopass :: write_init
    generic, public :: write_history=>write_image, write_hess, &
                       write_perturb, write_data,  &
                       write_nldata, write_nlshot, write_resd
    procedure, nopass, private :: write_image
    procedure, nopass, private :: write_hess
    procedure, nopass, private :: write_perturb
    procedure, nopass, private :: write_data
    procedure, nopass, private :: write_nldata
    procedure, nopass, private :: write_nlshot
    procedure, nopass, private :: write_resd
  end type

  interface
    pure module subroutine write_init(par)
      !! This procedure only allocates the array "d"
      implicit none
      type(read_), intent(in) :: par
    end subroutine

    pure module subroutine write_image(par,image,eic,out_fn)
      !! This procedure writes the output for the SEP history files and writes
      !! the migrated images to the corresponding binary files. It currently
      !! does not write the SEP headers.
      implicit none
      type(read_), intent(in) :: par
      type(rtm_image), intent(in) :: image
      logical, intent(in) :: eic
      character(len=*), optional, intent(in) :: out_fn
    end subroutine

    pure module subroutine write_hess(par,image,Hgn,nz,x1,x2,nlag1,nlag2)
      !! This procedure writes the output for the SEP history files and writes
      !! wemva perturbations to the corresponding binary files. It currently
      !! does not write the SEP headers.
      implicit none
      integer, intent(in) :: nz, x1, x2, nlag1, nlag2
      type(read_), intent(in) :: par
      type(rtm_image), intent(inout) :: image
      real, intent(in) :: Hgn(nz,x1:x2,-nlag1:nlag1,-nlag2:nlag2)
    end subroutine

    pure module subroutine write_perturb(par,image,wemva)
      !! This procedure writes the output for the SEP history files and writes
      !! wemva perturbations to the corresponding binary files. It currently
      !! does not write the SEP headers.
      implicit none
      type(read_), intent(in) :: par
      type(rtm_image), intent(inout) :: image
      type(wemva_foo), intent(in) :: wemva
    end subroutine

    pure module subroutine write_data(par,born,jsources)
      !! This procedure writes the output for the SEP history files and writes
      !! the Born modeled data to the corresponding binary files. It currently
      !! does not write the SEP headers.
      implicit none
      type(read_), intent(in) :: par
      type(born_foo), intent(in) :: born
      integer, intent(in) :: jsources
    end subroutine

    pure module subroutine write_nldata(par,born,nonlin,jsources)
      !! This procedure writes the output for the SEP history files and writes
      !! the non-linear modeled data to the corresponding binary files. It
      !! currently does not write the SEP headers.
      implicit none
      type(read_), intent(in) :: par
      type(born_foo), intent(inout) :: born
      type(nonlin_foo), intent(in) :: nonlin
      integer, intent(in) :: jsources
    end subroutine

    pure module subroutine write_nlshot(par,nonlin,jsources)
      !! This procedure writes the output for the SEP history files and writes
      !! the Born modeled data to the corresponding binary files. It currently
      !! does not write the SEP headers.
      implicit none
      type(read_), intent(in) :: par
      type(nonlin_foo), intent(in) :: nonlin
      integer, intent(in) :: jsources
    end subroutine

    pure module subroutine write_resd(par,born,jsources,lsq)
      !! This procedure writes the output for the SEP history files and writes
      !! the modeled residue to the corresponding binary files. It currently
      !! does not write the SEP headers.
      implicit none
      type(read_), intent(in) :: par
      type(born_foo), intent(in) :: born
      integer, intent(in) :: jsources
      logical, intent(in) :: lsq
    end subroutine

    pure module subroutine put_headers_data(par,sh_fn,jsources)
      !! This procedure puts header in modeled data
      implicit none
      type(read_), intent(in) :: par 
      character(len=*), intent(in) :: sh_fn
      integer, optional, intent(in) :: jsources
    end subroutine
  
  end interface 
end module
