submodule(write_mod) write_mod_implementation
  implicit none
contains
  module procedure write_init
    allocate(d(par%ntData), stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'write_init: could not allocate array "d"'
    end if
    if(par%projpref(1:4)/='none') then
      test: block 
        integer :: jsources, nchar
        character(3) :: suff
        nchar=len_trim(adjustl(par%projpref))
        allocate(character(nchar+4) :: shot_fn(par%nsources))
        do jsources=1,par%nsources
        shot_fn(jsources)(1:nchar-1)=trim(adjustl(par%projpref))
    
        write(suff,"(I3.3)") jsources
        shot_fn(jsources)(nchar:2+nchar)=suff
        shot_fn(jsources)(nchar+3:nchar+4)=".H"
        write(0,*) 'shot_fn: ',shot_fn(jsources)
        end do
      end block test
    end if
  end procedure

  module procedure write_image
    associate(o1=>par%o1,o2=>par%o2,d2=>par%d2,d3=>par%d3,unit1=>par%unit1,unit2=>par%unit2)
    associate(dz=>par%dz,dx=>par%dx,nz=>par%nz,nx=>par%nx,nkz=>par%nkz,nkgx=>par%nkgx)
    associate(first_cdp=>par%first_cdp,last_cdp=>par%last_cdp,nsources=>par%nsources,ntSnap=>par%ntSnap)
    associate(pad_=>par%pad_,zpad=>par%zpad,snapsht=>par%snapsht,one_src=>par%one_src)
    if(snapsht) then
      !! Write history files for snapshot files
      call to_history('o1',(1-pad_-zpad)*dz,"snapSrc")
      call to_history('o2',(first_cdp-pad_)*dx,"snapSrc")
      call to_history('o3',0,"snapSrc")
      call to_history('d1',dz,"snapSrc")
      call to_history('d2',dx,"snapSrc")
      call to_history('d3',1,"snapSrc")
      call to_history('n1',nz+2*pad_+zpad,"snapSrc")
      call to_history('n2',nx+2*pad_,"snapSrc")
      call to_history('n3',ntSnap,"snapSrc")
      call to_history('o1',(1-pad_-zpad)*dz,"snapRcv")
      call to_history('o2',(first_cdp-pad_)*dx,"snapRcv")
      call to_history('o3',0,"snapRcv")
      call to_history('d1',dz,"snapRcv")
      call to_history('d2',dx,"snapRcv")
      call to_history('d3',1,"snapRcv")
      call to_history('n1',nz+2*pad_+zpad,"snapRcv")
      call to_history('n2',nx+2*pad_,"snapRcv")
      call to_history('n3',ntSnap,"snapRcv")
    end if

    if(present(out_fn)) then
      allocate(out_,source=out_fn)
    else 
      out_="CIC"
    end if
    write(0,*) 'Writing history files'
    call to_history("title",'RTM shot CIC',out_)
    call to_history("n1",nz,  out_)
    call to_history("n2",nx,  out_)
    call to_history("d1",dz,  out_)
    call to_history("d2",dx,  out_)
    call to_history("dz",dz,  out_)
    call to_history("dcdp",dx,out_)
    call to_history("dline",1,out_)
    call to_history("o1",o1,  out_)
    call to_history("o2",o2,  out_)
    call to_history("label1",'depth',out_)
    call to_history("label2",'rec no.',out_)
    call to_history("unit1",unit1,out_)
    call to_history("unit2",unit2,out_)
    call to_history("esize",4,out_)
    
    if(par%eic) then 
      call to_history("title",'RTM shot EIC tomog.',"EIC1")
      call to_history("n1",nz,"EIC1")
      call to_history("n2",nx,"EIC1")
      call to_history("d1",dz,"EIC1")
      call to_history("d2",dx,"EIC1")
      call to_history("dz",dz,"EIC1")
      call to_history("dcdp",dx,"EIC1")
      call to_history("dline",1,"EIC1")
      call to_history("o1",o1,"EIC1")
      call to_history("o2",o2,"EIC1")
      call to_history("label1",'depth',"EIC1")
      call to_history("label2",'rec no.',"EIC1")
      call to_history("unit1",unit1,"EIC1")
      call to_history("unit2",unit2,"EIC1")
      call to_history("esize",4,"EIC1")
      
      call to_history("title",'RTM shot EIC refl.',"EIC2")
      call to_history("n1",nz,"EIC2")
      call to_history("n2",nx,"EIC2")
      call to_history("d1",dz,"EIC2")
      call to_history("d2",dx,"EIC2")
      call to_history("dz",dz,"EIC2")
      call to_history("dcdp",dx,"EIC2")
      call to_history("dline",1,"EIC2")
      call to_history("o1",o1,"EIC2")
      call to_history("o2",o2,"EIC2")
      call to_history("label1",'depth',"EIC2")
      call to_history("label2",'rec no.',"EIC2")
      call to_history("unit1",unit1,"EIC2")
      call to_history("unit2",unit2,"EIC2")
      call to_history("esize",4,"EIC2")
    end if
    
    call to_history('n1',nkz,"vrandom")
    call to_history('n2',nkgx,"vrandom")
    call to_history('n3',nsources,"vrandom")
    call to_history('o1',(1-pad_-zpad)*dz,"vrandom")
    call to_history('o2',(1-pad_)*dx,"vrandom")
    call to_history('o3',1,"vrandom")
    call to_history('d1',dz,"vrandom")
    call to_history('d2',dx,"vrandom")
    call to_history('d3',1,"vrandom")

    !! Write the header files of migrated shots (disabled for debugging purposes)
    !write(0,*) 'Writing headers'
    !call write_headers(cdp_no_key,x_key,  &
    !       n_keys_snap,sx_key_snap,sy_key_snap,gx_key_snap,gy_key_snap,  &
    !       hoff_key_snap,mute_key_snap,muts_key_snap,cdp_no_key_snap,line_no_key_snap,  &
    !       nt,ntraces,nz,nx,x,sx,gx,first_cdp,last_cdp,&
    !       z,wf_only)
    !* I use this to generate an empty hff file
    !call auxout("out")
    !if(sep_put_number_keys("in",n_keys_vel) /= 0) &
    !  write(0,*) 'Error, cannot write number of keys on the hff (.H@@) output file!'

!    if(.not. one_src) then
      !! Obtain amplitudes, non-zeros samples
      min_amplitude = 99999.0  
      max_amplitude = -99999.0
      n_nonzero_samples = 0
      associate(z1=>1,z2=>par%nz,x1=>par%first_cdp,x2=>last_cdp,I=>image%ICorr)
        !! Test for NaNs in ICorr
        call assert(.not.(any(isnan(I))),"write_image: NaNs in ICorr!")
        correlation_image: block
          integer :: jx, jz
          do jx = x1,x2
            do jz = z1,z2
              if(min_amplitude > I(jz,jx)) min_amplitude = I(jz,jx)
              if(max_amplitude < I(jz,jx)) max_amplitude = I(jz,jx)
              if(I(jz,jx) /= 0.) then
                amp_rms = amp_rms+I(jz,jx)*I(jz,jx)
                amp_mean = amp_mean+I(jz,jx)
                n_nonzero_samples = n_nonzero_samples+1
              end if
            end do
            !! Write image to binary file
            call sep_write(I(:,jx),out_)
          end do
          deallocate(out_)
        end block correlation_image
      end associate

      !! Test for zero outcome in ICorr
      call assert(n_nonzero_samples > 0, "write_image: ICorr is made of zeros!")
 
      amp_rms = SQRT(amp_rms/n_nonzero_samples)
      amp_mean = amp_mean/n_nonzero_samples
      call to_history("rms_amplitude",amp_rms,'out')
      call to_history("mean_amplitude",amp_mean,'out')
      call to_history("min_amplitude",min_amplitude,'out')
      call to_history("max_amplitude",max_amplitude,'out')
      call to_history("nonzero_samples",n_nonzero_samples,'out')
      write(0,*) '************** CIC Stack section:*******'
      write(0,*) 'Non-zero samples',n_nonzero_samples
      write(0,*) 'rms_amplitude',amp_rms
      write(0,*) 'mean_amplitude',amp_mean
      write(0,*) 'max_amplitude',max_amplitude
      write(0,*) 'min_amplitude',min_amplitude
      write(0,*) '****************************************'
    
      if(eic) then
        min_amplitude = 99999.0  
        max_amplitude = -99999.0
        n_nonzero_samples = 0
        associate(z1=>1,z2=>par%nz,x1=>par%first_cdp,x2=>last_cdp,I=>image%ITomo)
  
          !! Test for NaNs in ITomo
          call assert(.not.(any(isnan(I))),"write_image: NaNs in ITomo!")
  
          energy_image1: block
            integer :: jx, jz
            do jx = x1,x2
              do jz = z1,z2
                if(min_amplitude > I(jz,jx)) min_amplitude = I(jz,jx)
                if(max_amplitude < I(jz,jx)) max_amplitude = I(jz,jx)
                if(I(jz,jx) /= 0) then
                  amp_rms = amp_rms+I(jz,jx)*I(jz,jx)
                  amp_mean = amp_mean+I(jz,jx)
                  n_nonzero_samples = n_nonzero_samples+1
                end if
              end do
              call sep_write(I(:,jx),"EIC1")
            end do
          end block energy_image1
        end associate
       
        !! Test for zero outcome in ITomo
        call assert(n_nonzero_samples > 0, "write_image: ITomo is made of zeros!")
  
        amp_rms = SQRT(amp_rms/n_nonzero_samples)
        amp_mean = amp_mean/n_nonzero_samples
        call to_history("rms_amplitude",amp_rms)
        call to_history("mean_amplitude",amp_mean)
        call to_history("min_amplitude",min_amplitude)
        call to_history("max_amplitude",max_amplitude)
        call to_history("nonzero_samples",n_nonzero_samples)
        write(0,*) '****** EIC1 (tomog. comp.) Stack section:*******'
        write(0,*) 'Non-zero samples',n_nonzero_samples
        write(0,*) 'rms_amplitude',amp_rms
        write(0,*) 'mean_amplitude',amp_mean
        write(0,*) 'max_amplitude',max_amplitude
        write(0,*) 'min_amplitude',min_amplitude
        write(0,*) '***************************8********************'
      
        n_nonzero_samples = 0
        associate(z1=>1,z2=>par%nz,x1=>par%first_cdp,x2=>last_cdp,I=>image%IRefl)
          !! Test for NaNs in IRefl
          call assert(.not.(any(isnan(I))),"write_image: NaNs in IRefl!")
  
          energy_image2: block
            integer :: jx, jz
            do jx = x1,x2
              do jz = z1,z2
                if(min_amplitude > I(jz,jx)) min_amplitude = I(jz,jx)
                if(max_amplitude < I(jz,jx)) max_amplitude = I(jz,jx)
                if(I(jz,jx) /= 0) then
                  amp_rms = amp_rms+I(jz,jx)*I(jz,jx)
                  amp_mean = amp_mean+I(jz,jx)
                  n_nonzero_samples = n_nonzero_samples+1
                end if
              end do
              call sep_write(I(:,jx),"EIC2")
            end do
          end block energy_image2
        end associate
  
        !! Test for zero outcome in IRefl
        call assert(n_nonzero_samples > 0, "write_image: IRefl is made of zeros!")
  
        amp_rms = SQRT(amp_rms/n_nonzero_samples)
        amp_mean = amp_mean/n_nonzero_samples
        call to_history("rms_amplitude",amp_rms)
        call to_history("mean_amplitude",amp_mean)
        call to_history("min_amplitude",min_amplitude)
        call to_history("max_amplitude",max_amplitude)
        call to_history("nonzero_samples",n_nonzero_samples)
        write(0,*) '****** EIC2 (refl. comp.) Stack section:*******'
        write(0,*) 'Non-zero samples',n_nonzero_samples
        write(0,*) 'rms_amplitude',amp_rms
        write(0,*) 'mean_amplitude',amp_mean
        write(0,*) 'max_amplitude',max_amplitude
        write(0,*) 'min_amplitude',min_amplitude
        write(0,*) '***********************************************'
    end if
    end associate
    end associate
    end associate
    end associate
  end procedure

  module procedure write_hess
    associate(o1=>par%o1,o2=>par%o2,d2=>par%d2,unit1=>par%unit1,unit2=>par%unit2)
    associate(dz=>par%dz,dx=>par%dx,nz=>par%nz,nx=>par%nx)
      write(0,*) 'Writing history files'
      call to_history("title",'Gauss-Newton Hessian',"hess")
      call to_history("n1",nz,  "hess")
      call to_history("n2",nx,  "hess")
      call to_history("n3",2*nlag1+1,  "hess")
      call to_history("n4",2*nlag2+1,  "hess")
      call to_history("d1",dz,  "hess")
      call to_history("d2",dx,  "hess")
      call to_history("d3",1,  "hess")
      call to_history("d4",1,  "hess")
      call to_history("dz",dz,  "hess")
      call to_history("dcdp",dx,"hess")
      call to_history("dline",1,"hess")
      call to_history("o1",o1,  "hess")
      call to_history("o2",o2,  "hess")
      call to_history("o3",-nlag1,  "hess")
      call to_history("o4",-nlag2,  "hess")
      call to_history("label1",'depth',"hess")
      call to_history("label2",'rec no.',"hess")
      call to_history("label3",'lag1 ',"hess")
      call to_history("label4",'lag2 ',"hess")
      call to_history("unit1",unit1,"hess")
      call to_history("unit2",unit2,"hess")
      call to_history("unit3",'none',"hess")
      call to_history("unit4",'none',"hess")
      call to_history("esize",4,"hess")
    end associate
    end associate

    !! Obtain amplitudes, non-zeros samples
    min_amplitude = 99999.0  
    max_amplitude = -99999.0
    n_nonzero_samples = 0
    associate(z1=>1,z2=>par%nz,x1=>par%first_cdp,x2=>par%last_cdp)
      !! Test for NaNs in ICorr
      call assert(.not.(any(isnan(Hgn))),"write_hess: NaNs in Hgn!")
      correlation_image: block
        integer :: jx, jz, jlag1, jlag2
        do jlag2=-nlag2,nlag2
          do jlag1=-nlag1,nlag1
            do jx = x1,x2
              do jz = z1,z2
                if(min_amplitude > Hgn(jz,jx,jlag1,jlag2))&
                   min_amplitude = Hgn(jz,jx,jlag1,jlag2)
                if(max_amplitude < Hgn(jz,jx,jlag1,jlag2))&
                   max_amplitude = Hgn(jz,jx,jlag1,jlag2)
                if(Hgn(jz,jx,jlag1,jlag2) /= 0.) then
                  amp_rms = amp_rms+Hgn(jz,jx,jlag1,jlag2)*Hgn(jz,jx,jlag1,jlag2)
                  amp_mean = amp_mean+Hgn(jz,jx,jlag1,jlag2)
                  n_nonzero_samples = n_nonzero_samples+1
                end if
              end do
              !! Write image to binary file
              call sep_write(Hgn(:,jx,jlag1,jlag2),"hess")
            end do
          end do
        end do
      end block correlation_image
    end associate

    !! Test for zero outcome in ICorr
    call assert(n_nonzero_samples > 0, "write_hess: Hgn is made of zeros!")

    amp_rms = SQRT(amp_rms/n_nonzero_samples)
    amp_mean = amp_mean/n_nonzero_samples
    call to_history("rms_amplitude",amp_rms,"hess")
    call to_history("mean_amplitude",amp_mean,"hess")
    call to_history("min_amplitude",min_amplitude,"hess")
    call to_history("max_amplitude",max_amplitude,"hess")
    call to_history("nonzero_samples",n_nonzero_samples,"hess")
    write(0,*) '************** CIC Stack section:*******'
    write(0,*) 'Non-zero samples',n_nonzero_samples
    write(0,*) 'rms_amplitude',amp_rms
    write(0,*) 'mean_amplitude',amp_mean
    write(0,*) 'max_amplitude',max_amplitude
    write(0,*) 'min_amplitude',min_amplitude
    write(0,*) '****************************************'
  end procedure

  module procedure write_perturb
    allocate(image%ICorr,source=wemva%IPert)
    call write_image(par,image,.false.,"wemva")
    associate(dz=>par%dz,dx=>par%dx,nz=>par%nz,nx=>par%nx,nkz=>par%nkz,nkgx=>par%nkgx,nsources=>par%nsources)
    associate(pad_=>par%pad_,zpad=>par%zpad,snapsht=>par%snapsht,one_src=>par%one_src,ntSnap=>par%ntSnap)
      if(snapsht) then
        !! Write history files for snapshot files
        call to_history('o1',(1-pad_-zpad)*dz,"snapScat")
        call to_history('o2',(par%first_cdp-pad_)*dx,"snapScat")
        call to_history('o3',0,"snapScat")
        call to_history('d1',dz,"snapScat")
        call to_history('d2',dx,"snapScat")
        call to_history('d3',1,"snapScat")
        call to_history('n1',nz+2*pad_+zpad,"snapScat")
        call to_history('n2',nx+2*pad_,"snapScat")
        call to_history('n3',ntSnap,"snapScat")
      end if
    end associate
    end associate
  end procedure

  module procedure write_nldata
    if(.not. allocated(born%dsynt)) allocate(born%dsynt,mold=nonlin%dsynt)
    born%dsynt=nonlin%dsynt
    data_fn="out"
    call write_data(par,born,jsources)
    call to_history("title",'Non-linear modeled shot')
  end procedure

  module procedure write_data
    associate(o1=>par%o1,o2=>par%first_geof,d2=>par%d2,o3=>1,unit1=>par%unit1,unit2=>par%unit2)
    associate(dz=>par%dz,dx=>par%dx,nz=>par%nz,nx=>par%nx,nkz=>par%nkz,nkgx=>par%nkgx,nsources=>par%nsources)
    associate(dtData=>par%dtData,dsx=>par%dsx,dgx=>par%dgx,ntData=>par%ntData,ngx=>par%ngx)
    associate(pad_=>par%pad_,zpad=>par%zpad,snapsht=>par%snapsht,one_src=>par%one_src,ntSnap=>par%ntSnap)
      call to_history("title",'Born modeled shot',data_fn)
      call to_history("n1",ntData,                data_fn)
      call to_history("n2",ngx*nsources,          data_fn)
      call to_history("n3",1,                     data_fn)
      call to_history("d1",dtData,                data_fn)
      call to_history("d2",1,                     data_fn)
      call to_history("d3",1.,                    data_fn)
      call to_history("dt",dtData,                data_fn)
      call to_history("dcdp",dgx,                 data_fn)
      call to_history("dline",1,                  data_fn)
      call to_history("o1",0.,                    data_fn)
      call to_history("o2",o2,                    data_fn)
      call to_history("o3",o3,                    data_fn)
      call to_history("label1",'twt',             data_fn)
      call to_history("label2",'rec no.',         data_fn)
      call to_history("label3",' ',               data_fn)
      call to_history("unit1",unit1,              data_fn)
      call to_history("unit2",unit2,              data_fn)
      call to_history("unit3",'none',             data_fn)
      call to_history("esize",4,                  data_fn)

      call to_history('n1',nkz,"vrandom")
      call to_history('n2',nkgx,"vrandom")
      call to_history('n3',nsources,"vrandom")
      call to_history('o1',(1-pad_-zpad)*dz,"vrandom")
      call to_history('o2',par%sgx(1-par%napL,jsources)-pad_*dx,"vrandom")
      call to_history('o3',1,"vrandom")
      call to_history('d1',dz,"vrandom")
      call to_history('d2',dx,"vrandom")
      call to_history('d3',1,"vrandom")

      call to_history('n1',nkz,"rrandom")
      call to_history('n2',nkgx,"rrandom")
      call to_history('n3',nsources,"rrandom")
      call to_history('o1',(1-pad_-zpad)*dz,"rrandom")
      call to_history('o2',par%sgx(1-par%napL,jsources)-pad_*dx,"rrandom")
      call to_history('o3',1,"rrandom")
      call to_history('d1',dz,"rrandom")
      call to_history('d2',dx,"rrandom")
      call to_history('d3',1,"rrandom")

      if(snapsht) then
        !! Write history files for snapshot files
        call to_history('o1',(1-pad_-zpad)*dz,"snapSrc")
        call to_history('o2',(par%first_cdp-pad_)*dx,"snapSrc")
        call to_history('o3',0,"snapSrc")
        call to_history('d1',dz,"snapSrc")
        call to_history('d2',dx,"snapSrc")
        call to_history('d3',1,"snapSrc")
        call to_history('n1',nz+2*pad_+zpad,"snapSrc")
        call to_history('n2',nx+2*pad_,"snapSrc")
        call to_history('n3',ntSnap,"snapSrc")
        call to_history('o1',(1-pad_-zpad)*dz,"snapScat")
        call to_history('o2',(par%first_cdp-pad_)*dx,"snapScat")
        call to_history('o3',0,"snapScat")
        call to_history('d1',dz,"snapScat")
        call to_history('d2',dx,"snapScat")
        call to_history('d3',1,"snapScat")
        call to_history('n1',nz+2*pad_+zpad,"snapScat")
        call to_history('n2',nx+2*pad_,"snapScat")
        call to_history('n3',ntSnap,"snapScat")
      end if
    end associate
    end associate
    end associate
    end associate

    !! Obtain amplitudes, non-zeros samples
    min_amplitude = 99999.0  
    max_amplitude = -99999.0
    n_nonzero_samples = 0

    !! Test for NaNs in dsynt
    call assert(.not.(any(isnan(born%dsynt))),"write_data(1): NaNs in dsynt!")
    call assert(norm2(born%dsynt)/=0.,'write_data(2): dsynt is made of zeros!')

    !! Save modeled data
    save_data: block
      integer :: jt, jgx
      real :: tIntFact
      tIntFact=par%dtData/par%dt
      do jgx = 1,par%ngx
        call interp(par%ntData,tIntFact,par%nt,d,born%dsynt(:,jgx),.true.)
        do jt = 1,par%ntData
          if(min_amplitude > d(jt)) min_amplitude = d(jt)
          if(max_amplitude < d(jt)) max_amplitude = d(jt)
          if(d(jt) /= 0) then
            amp_rms = amp_rms+d(jt)*d(jt)
            amp_mean = amp_mean+d(jt)
            n_nonzero_samples = n_nonzero_samples+1
          end if
        end do
!        !! Write data to binary file
!        call sep_write(d(:))

        !! Writes  data traces
        previous_trace = (jsources-1)*par%ngx+jgx-1
        if(sseek_block(data_fn,previous_trace,4*par%ntData,0) < 0) &
        call seperr('write_data: Have a trouble seeking trace')
        if(srite(data_fn,d,4*par%ntData) /= 4*par%ntData) &
        call seperr('write_data: Have a trouble writing trace')
      end do
    end block save_data

    !! Test for zero outcome in dsynt
    call assert(n_nonzero_samples > 0, "write_data(3): dsynt is made of zeros!")
 
    amp_rms = SQRT(amp_rms/n_nonzero_samples)
    amp_mean = amp_mean/n_nonzero_samples
    call to_history("rms_amplitude",amp_rms,            data_fn)
    call to_history("mean_amplitude",amp_mean,          data_fn)
    call to_history("min_amplitude",min_amplitude,      data_fn)
    call to_history("max_amplitude",max_amplitude,      data_fn)
    call to_history("nonzero_samples",n_nonzero_samples,data_fn)
    write(0,*) 'Non-zero samples',n_nonzero_samples
    write(0,*) 'rms_amplitude',amp_rms
    write(0,*) 'mean_amplitude',amp_mean
    write(0,*) 'max_amplitude',max_amplitude
    write(0,*) 'min_amplitude',min_amplitude
  end procedure

  module procedure write_nlshot
    associate(o1=>par%o1,o2=>par%first_geof,d2=>par%d2,o3=>1,unit1=>par%unit1,unit2=>par%unit2)
    associate(dz=>par%dz,dx=>par%dx,nz=>par%nz,nx=>par%nx,nkz=>par%nkz,nkgx=>par%nkgx,nsources=>par%nsources)
    associate(dtData=>par%dtData,dsx=>par%dsx,dgx=>par%dgx,ntData=>par%ntData,ngx=>par%ngx)
    associate(pad_=>par%pad_,zpad=>par%zpad,snapsht=>par%snapsht,one_src=>par%one_src,ntSnap=>par%ntSnap)

      write(0,*) 'write_nlshot: Writing modeled shot: ',shot_fn(jsources)

      call to_history("title",'Born modeled shot',shot_fn(jsources))
      call to_history("n1",ntData,shot_fn(jsources))
      call to_history("n2",ngx,shot_fn(jsources))
      call to_history("n3",1,shot_fn(jsources))
      call to_history("d1",dtData,shot_fn(jsources))
      call to_history("d2",1,shot_fn(jsources))
      call to_history("d3",1.,shot_fn(jsources))
      call to_history("dt",dtData,shot_fn(jsources))
      call to_history("dcdp",dgx,shot_fn(jsources))
      call to_history("dline",1,shot_fn(jsources))
      call to_history("o1",0.,shot_fn(jsources))
      call to_history("o2",o2,shot_fn(jsources))
      call to_history("o3",o3,shot_fn(jsources))
      call to_history("label1",'twt',shot_fn(jsources))
      call to_history("label2",'rec no.',shot_fn(jsources))
      call to_history("label3",' ',shot_fn(jsources))
      call to_history("unit1",unit1,shot_fn(jsources))
      call to_history("unit2",unit2,shot_fn(jsources))
      call to_history("unit3",'none',shot_fn(jsources))
      call to_history("esize",4,shot_fn(jsources))

      call put_headers_data(par,shot_fn(jsources),jsources)

      if(snapsht) then
        !! Write history files for snapshot files
        call to_history('o1',(1-pad_-zpad)*dz,"snapSrc")
        call to_history('o2',(par%first_cdp-pad_)*dx,"snapSrc")
        call to_history('o3',0,"snapSrc")
        call to_history('d1',dz,"snapSrc")
        call to_history('d2',dx,"snapSrc")
        call to_history('d3',1,"snapSrc")
        call to_history('n1',nz+2*pad_+zpad,"snapSrc")
        call to_history('n2',nx+2*pad_,"snapSrc")
        call to_history('n3',ntSnap,"snapSrc")
        call to_history('o1',(1-pad_-zpad)*dz,"snapScat")
        call to_history('o2',(par%first_cdp-pad_)*dx,"snapScat")
        call to_history('o3',0,"snapScat")
        call to_history('d1',dz,"snapScat")
        call to_history('d2',dx,"snapScat")
        call to_history('d3',1,"snapScat")
        call to_history('n1',nz+2*pad_+zpad,"snapScat")
        call to_history('n2',nx+2*pad_,"snapScat")
        call to_history('n3',ntSnap,"snapScat")
      end if
      call to_history('n1',nkz,"vrandom")
      call to_history('n2',nkgx,"vrandom")
      call to_history('n3',nsources,"vrandom")
      call to_history('o1',(1-pad_-zpad)*dz,"vrandom")
      call to_history('o2',par%sgx(1-par%napL,jsources)-pad_*dx,"vrandom")
      call to_history('o3',1,"vrandom")
      call to_history('d1',dz,"vrandom")
      call to_history('d2',dx,"vrandom")
      call to_history('d3',1,"vrandom")

    end associate
    end associate
    end associate
    end associate

    !! Obtain amplitudes, non-zeros samples
    min_amplitude = 99999.0  
    max_amplitude = -99999.0
    n_nonzero_samples = 0

    !! Test for NaNs in dsynt
    call assert(.not.(any(isnan(nonlin%dsynt))),"write_nlshot(1): NaNs in dsynt!")
    call assert(norm2(nonlin%dsynt)/=0.,'write_nlshot(2): dsynt is made of zeros!')

    !! Save modeled data
    save_data: block
      integer :: jt, jgx
      real :: tIntFact
      tIntFact=par%dtData/par%dt
      do jgx = 1,par%ngx
        call interp(par%ntData,tIntFact,par%nt,d,nonlin%dsynt(:,jgx),.true.)
        do jt = 1,par%ntData
          if(min_amplitude > d(jt)) min_amplitude = d(jt)
          if(max_amplitude < d(jt)) max_amplitude = d(jt)
          if(d(jt) /= 0) then
            amp_rms = amp_rms+d(jt)*d(jt)
            amp_mean = amp_mean+d(jt)
            n_nonzero_samples = n_nonzero_samples+1
          end if
        end do

        !! Writes  data traces
!        previous_trace = (jsources-1)*par%ngx+jgx-1
        previous_trace = jgx-1
        if(sseek_block(shot_fn(jsources),previous_trace,4*par%ntData,0) < 0) &
        call seperr('write_nldata: Have a trouble seeking trace')
        if(srite(shot_fn(jsources),d,4*par%ntData) /= 4*par%ntData) &
        call seperr('write_nldata: Have a trouble writing trace')
      end do
    end block save_data

    !! Test for zero outcome in dsynt
    call assert(n_nonzero_samples > 0, "write_data(2): dsynt is made of zeros!")
 
    amp_rms = SQRT(amp_rms/n_nonzero_samples)
    amp_mean = amp_mean/n_nonzero_samples
    call to_history("rms_amplitude",amp_rms,shot_fn(jsources))
    call to_history("mean_amplitude",amp_mean,shot_fn(jsources))
    call to_history("min_amplitude",min_amplitude,shot_fn(jsources))
    call to_history("max_amplitude",max_amplitude,shot_fn(jsources))
    call to_history("nonzero_samples",n_nonzero_samples,shot_fn(jsources))
    write(0,*) 'Non-zero samples',n_nonzero_samples
    write(0,*) 'rms_amplitude',amp_rms
    write(0,*) 'mean_amplitude',amp_mean
    write(0,*) 'max_amplitude',max_amplitude
    write(0,*) 'min_amplitude',min_amplitude
  end procedure

  module procedure write_resd
    associate(o1=>par%o1,o2=>par%first_geof,d2=>par%d2,o3=>1,unit1=>par%unit1,unit2=>par%unit2)
    associate(dz=>par%dz,dx=>par%dx,nz=>par%nz,nx=>par%nx,nkz=>par%nkz,nkgx=>par%nkgx,nsources=>par%nsources)
    associate(dtData=>par%dtData,dsx=>par%dsx,dgx=>par%dgx,ntData=>par%ntData,ngx=>par%ngx)
    associate(pad_=>par%pad_,zpad=>par%zpad,one_src=>par%one_src)
      call to_history("title",'Residual','resd')
      call to_history("n1",ntData,'resd')
      call to_history("n2",ngx*nsources,'resd')
      call to_history("n3",1,'resd')
      call to_history("d1",dtData,'resd')
      call to_history("d2",1,'resd')
      call to_history("d3",1.,'resd')
      call to_history("dt",dtData,'resd')
      call to_history("dcdp",dgx,'resd')
      call to_history("dline",1,'resd')
      call to_history("o1",0.,'resd')
      call to_history("o2",o2,'resd')
      call to_history("o3",o3,'resd')
      call to_history("label1",'twt','resd')
      call to_history("label2",'rec no.','resd')
      call to_history("label3",' ','resd')
      call to_history("unit1",unit1,'resd')
      call to_history("unit2",unit2,'resd')
      call to_history("unit3",'none','resd')
      call to_history("esize",4,'resd')

    end associate
    end associate
    end associate
    end associate

    !! Obtain amplitudes, non-zeros samples
    min_amplitude = 99999.0  
    max_amplitude = -99999.0
    n_nonzero_samples = 0

    !! Test for NaNs in resd
    call assert(.not.(any(isnan(born%resd))),"write_resd(1): NaNs in resd!")
    call assert(norm2(born%resd)/=0.,'write_resd(2): resd is made of zeros!')

    !! Save residual
    save_resd: block
      integer :: jt, jgx
      real :: tIntFact
      tIntFact=par%dtData/par%dt
      do jgx = 1,par%ngx
        call interp(par%ntData,tIntFact,par%nt,d,born%resd(:,jgx),.true.)
        do jt = 1,par%ntData
          if(min_amplitude > d(jt)) min_amplitude = d(jt)
          if(max_amplitude < d(jt)) max_amplitude = d(jt)
          if(d(jt) /= 0) then
            amp_rms = amp_rms+d(jt)*d(jt)
            amp_mean = amp_mean+d(jt)
            n_nonzero_samples = n_nonzero_samples+1
          end if
        end do

        !! Writes resdidue traces
!        previous_trace = jgx-1
        previous_trace = (jsources-1)*par%ngx+jgx-1
        if(sseek_block("resd",previous_trace,4*par%ntData,0) < 0) &
        call seperr('write_resd: Have a trouble seeking trace')
        if(srite("resd",d,4*par%ntData) /= 4*par%ntData) &
        call seperr('write_resd: Have a trouble writing trace')
      end do
    end block save_resd

    !! Test for zero outcome in resd
    call assert(n_nonzero_samples > 0, "write_resd(3): resd is made of zeros!")
 
    amp_rms = SQRT(amp_rms/n_nonzero_samples)
    amp_mean = amp_mean/n_nonzero_samples
    call to_history("rms_amplitude",amp_rms,'resd')
    call to_history("mean_amplitude",amp_mean,'resd')
    call to_history("min_amplitude",min_amplitude,'resd')
    call to_history("max_amplitude",max_amplitude,'resd')
    call to_history("nonzero_samples",n_nonzero_samples,'resd')
    write(0,*) 'Non-zero samples',n_nonzero_samples
    write(0,*) 'rms_amplitude',amp_rms
    write(0,*) 'mean_amplitude',amp_mean
    write(0,*) 'max_amplitude',max_amplitude
    write(0,*) 'min_amplitude',min_amplitude
  end procedure

  module procedure put_headers_data
    associate(ntraces=>par%ntraces,ngx=>par%ngx)
!      dummy=sep_get_header_format_tag("datafile",tag_input)
      output_error=0
      write(0,*) ' '
      write(0,*) 'put_headers_data: Write headers on modeled data'
      if(sep_put_number_keys(sh_fn,5) /= 0) then
        write(0,*) 'Error, cannot put in number of keys from the hff (.H@@) file!'
        output_error=output_error+1
      end if
      if(sep_put_key(sh_fn,"sx","scalar_float","xdr_float",1) /= 0) then
        write(0,*) 'error! sx key index cannot be written in modeled data file!'
        output_error=output_error+1
      end if
      if(sep_put_key(sh_fn,"gx","scalar_float","xdr_float",2) /= 0) then
        write(0,*) 'error! gx key index cannot be written in modeled data file!'
        output_error=output_error+1
      end if
      if(sep_put_key(sh_fn,"sy","scalar_float","xdr_float",3) /= 0) then
        write(0,*) 'error! sy key index cannot be written in modeled data file!'
        output_error=output_error+1
      end if
      if(sep_put_key(sh_fn,"gy","scalar_float","xdr_float",4) /= 0) then
        write(0,*) 'error! gy key index cannot be written in modeled data file!'
        output_error=output_error+1
      end if
      if(sep_put_key(sh_fn,"scalco","scalar_float","xdr_float",5) /= 0) then
        write(0,*) 'error! scalco key cannot be writen in modeled data file!'
        output_error=output_error+1
      end if

      scalcof = par%scalco
      if(par%scalco < 0) then
        scalcof = -1./scalcof
      end if

      header_out = 0
      header: block
        integer :: first_trace, last_trace, nout
        if(present(jsources)) then
          allocate(header_out(5,ngx))
          nout=ngx
          first_trace=(jsources-1)*ngx+1
          last_trace = jsources*ngx
          header_out(1,:)=sx(first_trace:last_trace)/scalcof
          header_out(2,:)=gx(first_trace:last_trace)/scalcof
          header_out(3,:)=0
          header_out(4,:)=0
          header_out(5,:)=par%scalco
        else
          allocate(header_out(5,par%ntraces))
          nout=par%ntraces
          header_out(1,:)=sx/scalcof
          header_out(2,:)=gx/scalcof
          header_out(3,:)=0
          header_out(4,:)=0
          header_out(5,:)=par%scalco
        end if

        if(sep_put_data_axis_par(sh_fn,2,nout,0.,1.,"trace_number") /= 0) then
          call skull()
          error stop 'put_headers_data: could not write data_axis_par!'
        end if
        if(sep_put_header_axis_par(sh_fn,2,nout,0.,1.,"trace_number") /= 0) then
          call skull()
          error stop 'put_headers_data: could not write header_axis_par!'
        end if
!        if(sep_put_val_headers(sh_fn,1,ntraces,header_out) /= 0) then
        do jgx=1,nout
          if(sep_put_val_headers(sh_fn,jgx,1,1.*header_out(:,jgx)) /= 0) then
            call skull()
            error stop 'put_headers_data: could not write to output data headers!'
          end if
        end do
      end block header

      deallocate(header_out)
    end associate
  end procedure

end submodule
